#
# Copyright (C) 2008 Martijn Voncken <mvoncken@gmail.com>
#
# Basic plugin template created by:
# Copyright (C) 2008 Martijn Voncken <mvoncken@gmail.com>
# Copyright (C) 2007-2008 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

"""
torrent-label core plugin.
adds a status field for tracker.
"""
import logging
import re

import deluge.component as component
from deluge.configmanager import ConfigManager
from deluge.core.rpcserver import export
from deluge.plugins.pluginbase import CorePluginBase

log = logging.getLogger(__name__)

RE_VALID = re.compile(r'[a-z0-9_\-\.]*\Z')

KNOWN_STATES = ['Downloading', 'Seeding', 'Paused', 'Checking', 'Queued', 'Error']
STATE = 'state'
TRACKER = 'tracker'
KEYWORD = 'keyword'
LABEL = 'label'
CONFIG_DEFAULTS = {
    'torrent_labels': {},  # torrent_id:label_id
    'labels': {},  # label_id:{name:value}
}

CORE_OPTIONS = ['auto_add_trackers']

OPTIONS_DEFAULTS = {
    'apply_max': False,
    'max_download_speed': -1,
    'max_upload_speed': -1,
    'max_connections': -1,
    'max_upload_slots': -1,
    'prioritize_first_last': False,
    'apply_queue': False,
    'is_auto_managed': False,
    'stop_at_ratio': False,
    'stop_ratio': 2.0,
    'remove_at_ratio': False,
    'apply_move_completed': False,
    'move_completed': False,
    'move_completed_path': '',
    'auto_add': False,
    'auto_add_trackers': [],
}

NO_LABEL = 'No Label'


def check_input(cond, message):
    if not cond:
        raise Exception(message)


class Core(CorePluginBase):
    """
    self.labels = {label_id:label_options_dict}
    self.torrent_labels = {torrent_id:label_id}
    """

    def enable(self):
        log.info('*** Start Label plugin ***')
        self.plugin = component.get('CorePluginManager')
        self.plugin.register_status_field('label', self._status_get_label)

        # __init__
        core = component.get('Core')
        self.config = ConfigManager('label.conf', defaults=CONFIG_DEFAULTS)
        self.core_cfg = ConfigManager('core.conf')

        # reduce typing, assigning some values to self...
        self.torrents = core.torrentmanager.torrents
        self.labels = self.config['labels']
        self.torrent_labels = self.config['torrent_labels']

        self.clean_initial_config()

        component.get('EventManager').register_event_handler(
            'TorrentAddedEvent', self.post_torrent_add
        )
        component.get('EventManager').register_event_handler(
            'TorrentRemovedEvent', self.post_torrent_remove
        )

        # register tree:
        component.get('FilterManager').register_tree_field(
            'label', self.init_filter_dict
        )

        log.debug('Label plugin enabled..')

    def disable(self):
        self.plugin.deregister_status_field('label')
        component.get('FilterManager').deregister_tree_field('label')
        component.get('EventManager').deregister_event_handler(
            'TorrentAddedEvent', self.post_torrent_add
        )
        component.get('EventManager').deregister_event_handler(
            'TorrentRemovedEvent', self.post_torrent_remove
        )

    def update(self):
        pass

    def init_filter_dict(self):
        filter_dict = {label: 0 for label in self.labels}
        filter_dict['All'] = len(self.torrents)
        return filter_dict

    # Plugin hooks #
    def post_torrent_add(self, torrent_id, from_state):
        if from_state:
            return
        log.debug('post_torrent_add')
        torrent = self.torrents[torrent_id]

        for label_id, options in self.labels.items():
            if options['auto_add']:
                if self._has_auto_match(torrent, options):
                    self.set_torrent(torrent_id, label_id)
                    return

    def post_torrent_remove(self, torrent_id):
        log.debug('post_torrent_remove')
        if torrent_id in self.torrent_labels:
            del self.torrent_labels[torrent_id]

    # Utils #
    def clean_config(self):
        """remove invalid data from config-file"""
        for torrent_id, label_id in list(self.torrent_labels.items()):
            if (label_id not in self.labels) or (torrent_id not in self.torrents):
                log.debug('label: rm %s:%s', torrent_id, label_id)
                del self.torrent_labels[torrent_id]

    def clean_initial_config(self):
        """
        *add any new keys in OPTIONS_DEFAULTS
        *set all None values to default <-fix development config
        """
        log.debug(list(self.labels))
        for key in self.labels:
            options = dict(OPTIONS_DEFAULTS)
            options.update(self.labels[key])
            self.labels[key] = options

        for label, options in self.labels.items():
            for key, value in options.items():
                if value is None:
                    self.labels[label][key] = OPTIONS_DEFAULTS[key]

    def save_config(self):
        self.clean_config()
        self.config.save()

    @export
    def get_labels(self):
        return sorted(self.labels)

    # Labels:
    @export
    def add(self, label_id):
        """add a label
        see label_set_options for more options.
        """
        label_id = label_id.lower()
        check_input(
            RE_VALID.match(label_id), _('Invalid label, valid characters:[a-z0-9_-]')
        )
        check_input(label_id, _('Empty Label'))
        check_input(not (label_id in self.labels), _('Label already exists'))

        self.labels[label_id] = dict(OPTIONS_DEFAULTS)
        self.config.save()

    @export
    def remove(self, label_id):
        """remove a label"""
        check_input(label_id in self.labels, _('Unknown Label'))
        del self.labels[label_id]
        self.clean_config()
        self.config.save()

    def _set_torrent_options(self, torrent_id, label_id):
        options = self.labels[label_id]
        torrent = self.torrents[torrent_id]

        if not options['move_completed_path']:
            options['move_completed_path'] = ''  # no None.

        if options['apply_max']:
            torrent.set_max_download_speed(options['max_download_speed'])
            torrent.set_max_upload_speed(options['max_upload_speed'])
            torrent.set_max_connections(options['max_connections'])
            torrent.set_max_upload_slots(options['max_upload_slots'])
            torrent.set_prioritize_first_last_pieces(options['prioritize_first_last'])

        if options['apply_queue']:
            torrent.set_auto_managed(options['is_auto_managed'])
            torrent.set_stop_at_ratio(options['stop_at_ratio'])
            torrent.set_stop_ratio(options['stop_ratio'])
            torrent.set_remove_at_ratio(options['remove_at_ratio'])

        if options['apply_move_completed']:
            torrent.set_options(
                {
                    'move_completed': options['move_completed'],
                    'move_completed_path': options['move_completed_path'],
                }
            )

    def _unset_torrent_options(self, torrent_id, label_id):
        options = self.labels[label_id]
        torrent = self.torrents[torrent_id]

        if options['apply_max']:
            torrent.set_max_download_speed(
                self.core_cfg.config['max_download_speed_per_torrent']
            )
            torrent.set_max_upload_speed(
                self.core_cfg.config['max_upload_speed_per_torrent']
            )
            torrent.set_max_connections(
                self.core_cfg.config['max_connections_per_torrent']
            )
            torrent.set_max_upload_slots(
                self.core_cfg.config['max_upload_slots_per_torrent']
            )
            torrent.set_prioritize_first_last_pieces(
                self.core_cfg.config['prioritize_first_last_pieces']
            )

        if options['apply_queue']:
            torrent.set_auto_managed(self.core_cfg.config['auto_managed'])
            torrent.set_stop_at_ratio(self.core_cfg.config['stop_seed_at_ratio'])
            torrent.set_stop_ratio(self.core_cfg.config['stop_seed_ratio'])
            torrent.set_remove_at_ratio(self.core_cfg.config['remove_seed_at_ratio'])

        if options['apply_move_completed']:
            torrent.set_options(
                {
                    'move_completed': self.core_cfg.config['move_completed'],
                    'move_completed_path': self.core_cfg.config['move_completed_path'],
                }
            )

    def _has_auto_match(self, torrent, label_options):
        """match for auto_add fields"""
        for tracker_match in label_options['auto_add_trackers']:
            for tracker in torrent.trackers:
                if tracker_match in tracker['url']:
                    return True
        return False

    @export
    def set_options(self, label_id, options_dict):
        """update the label options

        options_dict :
        {"max_download_speed":float(),
            "max_upload_speed":float(),
            "max_connections":int(),
            "max_upload_slots":int(),
            #"prioritize_first_last":bool(),
            "apply_max":bool(),
            "move_completed_to":string() or None
        }
        """
        check_input(label_id in self.labels, _('Unknown Label'))
        for key in options_dict:
            if key not in OPTIONS_DEFAULTS:
                raise Exception('label: Invalid options_dict key:%s' % key)

        self.labels[label_id].update(options_dict)

        # apply
        for torrent_id, label in self.torrent_labels.items():
            if label_id == label and torrent_id in self.torrents:
                self._set_torrent_options(torrent_id, label_id)

        # auto add
        options = self.labels[label_id]
        if options['auto_add']:
            for torrent_id, torrent in self.torrents.items():
                if self._has_auto_match(torrent, options):
                    self.set_torrent(torrent_id, label_id)

        self.config.save()

    @export
    def get_options(self, label_id):
        """returns the label options"""
        return self.labels[label_id]

    @export
    def set_torrent(self, torrent_id, label_id):
        """
        assign a label to a torrent
        removes a label if the label_id parameter is empty.
        """
        if label_id == NO_LABEL:
            label_id = None

        check_input((not label_id) or (label_id in self.labels), _('Unknown Label'))
        check_input(torrent_id in self.torrents, _('Unknown Torrent'))

        if torrent_id in self.torrent_labels:
            self._unset_torrent_options(torrent_id, self.torrent_labels[torrent_id])
            del self.torrent_labels[torrent_id]
            self.clean_config()
        if label_id:
            self.torrent_labels[torrent_id] = label_id
            self._set_torrent_options(torrent_id, label_id)

        self.config.save()

    @export
    def get_config(self):
        """see : label_set_config"""
        return {
            key: self.config[key] for key in CORE_OPTIONS if key in self.config.config
        }

    @export
    def set_config(self, options):
        """global_options:"""
        if options:
            for key, value in options.items:
                if key in CORE_OPTIONS:
                    self.config[key] = value

            self.config.save()

    def _status_get_label(self, torrent_id):
        return self.torrent_labels.get(torrent_id) or ''



#
# Copyright (C) 2010 John Garland <johnnybg+deluge@gmail.com>
#
# Basic plugin template created by:
# Copyright (C) 2008 Martijn Voncken <mvoncken@gmail.com>
# Copyright (C) 2007-2009 Andrew Resch <andrewresch@gmail.com>
# Copyright (C) 2009 Damien Churchill <damoxc@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

from setuptools import find_packages, setup

__plugin_name__ = 'Toggle'
__author__ = 'John Garland'
__author_email__ = 'johnnybg+deluge@gmail.com'
__version__ = '0.4'
__url__ = 'http://deluge-torrent.org'
__license__ = 'GPLv3'
__description__ = 'Toggles the session'
__long_description__ = """"""
__pkg_data__ = {'deluge_' + __plugin_name__.lower(): ['data/*']}

setup(
    name=__plugin_name__,
    version=__version__,
    description=__description__,
    author=__author__,
    author_email=__author_email__,
    url=__url__,
    license=__license__,
    long_description=__long_description__ if __long_description__ else __description__,
    packages=find_packages(),
    package_data=__pkg_data__,
    entry_points="""
    [deluge.plugin.core]
    %s = deluge_%s:CorePlugin
    [deluge.plugin.gtk3ui]
    %s = deluge_%s:GtkUIPlugin
    [deluge.plugin.web]
    %s = deluge_%s:WebUIPlugin
    """
    % ((__plugin_name__, __plugin_name__.lower()) * 3),
)



#
# Copyright (C) 2009 GazpachoKing <chase.sterling@gmail.com>
# Copyright (C) 2011 Pedro Algarvio <pedro@algarvio.me>
#
# Basic plugin template created by:
# Copyright (C) 2008 Martijn Voncken <mvoncken@gmail.com>
# Copyright (C) 2007-2009 Andrew Resch <andrewresch@gmail.com>
# Copyright (C) 2009 Damien Churchill <damoxc@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging
import os
import shutil
from base64 import b64encode

from twisted.internet import reactor
from twisted.internet.defer import maybeDeferred
from twisted.internet.task import LoopingCall, deferLater
from twisted.python.failure import Failure

import deluge.component as component
import deluge.configmanager
from deluge._libtorrent import lt
from deluge.common import AUTH_LEVEL_ADMIN, is_magnet
from deluge.core.rpcserver import export
from deluge.error import AddTorrentError, InvalidTorrentError
from deluge.event import DelugeEvent
from deluge.plugins.pluginbase import CorePluginBase

log = logging.getLogger(__name__)


DEFAULT_PREFS = {'watchdirs': {}, 'next_id': 1}


OPTIONS_AVAILABLE = {  # option: builtin
    'enabled': False,
    'path': False,
    'append_extension': False,
    'copy_torrent': False,
    'delete_copy_torrent_toggle': False,
    'abspath': False,
    'download_location': True,
    'max_download_speed': True,
    'max_upload_speed': True,
    'max_connections': True,
    'max_upload_slots': True,
    'prioritize_first_last': True,
    'auto_managed': True,
    'stop_at_ratio': True,
    'stop_ratio': True,
    'remove_at_ratio': True,
    'move_completed': True,
    'move_completed_path': True,
    'label': False,
    'add_paused': True,
    'queue_to_top': False,
    'owner': True,
    'seed_mode': True,
}

MAX_NUM_ATTEMPTS = 10


class AutoaddOptionsChangedEvent(DelugeEvent):
    """Emitted when the options for the plugin are changed."""

    def __init__(self):
        pass


def check_input(cond, message):
    if not cond:
        raise Exception(message)


class Core(CorePluginBase):
    def enable(self):

        # reduce typing, assigning some values to self...
        self.config = deluge.configmanager.ConfigManager('autoadd.conf', DEFAULT_PREFS)
        self.config.run_converter((0, 1), 2, self.__migrate_config_1_to_2)
        self.config.save()
        self.watchdirs = self.config['watchdirs']

        self.rpcserver = component.get('RPCServer')
        component.get('EventManager').register_event_handler(
            'PreTorrentRemovedEvent', self.__on_pre_torrent_removed
        )

        # Dict of Filename:Attempts
        self.invalid_torrents = {}
        # Loopingcall timers for each enabled watchdir
        self.update_timers = {}
        deferLater(reactor, 5, self.enable_looping)

    def enable_looping(self):
        # Enable all looping calls for enabled watchdirs here
        for watchdir_id, watchdir in self.watchdirs.items():
            if watchdir['enabled']:
                self.enable_watchdir(watchdir_id)

    def disable(self):
        # disable all running looping calls
        component.get('EventManager').deregister_event_handler(
            'PreTorrentRemovedEvent', self.__on_pre_torrent_removed
        )
        for loopingcall in self.update_timers.values():
            loopingcall.stop()
        self.config.save()

    def update(self):
        pass

    @export
    def set_options(self, watchdir_id, options):
        """Update the options for a watch folder."""
        watchdir_id = str(watchdir_id)
        options = self._make_unicode(options)
        check_input(watchdir_id in self.watchdirs, _('Watch folder does not exist.'))
        if 'path' in options:
            options['abspath'] = os.path.abspath(options['path'])
            check_input(os.path.isdir(options['abspath']), _('Path does not exist.'))
            for w_id, w in self.watchdirs.items():
                if options['abspath'] == w['abspath'] and watchdir_id != w_id:
                    raise Exception('Path is already being watched.')
        for key in options:
            if key not in OPTIONS_AVAILABLE:
                if key not in [key2 + '_toggle' for key2 in OPTIONS_AVAILABLE]:
                    raise Exception('autoadd: Invalid options key:%s' % key)
        # disable the watch loop if it was active
        if watchdir_id in self.update_timers:
            self.disable_watchdir(watchdir_id)

        self.watchdirs[watchdir_id].update(options)
        # re-enable watch loop if appropriate
        if self.watchdirs[watchdir_id]['enabled']:
            self.enable_watchdir(watchdir_id)
        self.config.save()
        component.get('EventManager').emit(AutoaddOptionsChangedEvent())

    def load_torrent(self, filename, magnet):
        log.debug('Attempting to open %s for add.', filename)
        file_mode = 'r' if magnet else 'rb'
        try:
            with open(filename, file_mode) as _file:
                filedump = _file.read()
        except OSError as ex:
            log.warning('Unable to open %s: %s', filename, ex)
            raise ex

        if not filedump:
            raise EOFError('Torrent is 0 bytes!')

        # Get the info to see if any exceptions are raised
        if not magnet:
            decoded_torrent = lt.bdecode(filedump)
            if decoded_torrent is None:
                raise InvalidTorrentError('Torrent file failed decoding.')
            lt.torrent_info(decoded_torrent)

        return filedump

    def split_magnets(self, filename):
        log.debug('Attempting to open %s for splitting magnets.', filename)
        magnets = []
        try:
            with open(filename) as _file:
                magnets = list(filter(len, _file.read().splitlines()))
        except OSError as ex:
            log.warning('Unable to open %s: %s', filename, ex)

        if len(magnets) < 2:
            return []

        path = filename.rsplit(os.sep, 1)[0]
        for magnet in magnets:
            if not is_magnet(magnet):
                log.warning('Found line which is not a magnet: %s', magnet)
                continue

            for part in magnet.split('&'):
                if part.startswith('dn='):
                    name = part[3:].strip()
                    if name:
                        mname = os.sep.join([path, name + '.magnet'])
                        break
            else:
                short_hash = magnet.split('btih:')[1][:8]
                mname = '.'.join([os.path.splitext(filename)[0], short_hash, 'magnet'])

            try:
                with open(mname, 'w') as _mfile:
                    _mfile.write(magnet)
            except OSError as ex:
                log.warning('Unable to open %s: %s', mname, ex)
        return magnets

    def update_watchdir(self, watchdir_id):
        """Check the watch folder for new torrents to add."""
        log.trace('Updating watchdir id: %s', watchdir_id)
        watchdir_id = str(watchdir_id)
        watchdir = self.watchdirs[watchdir_id]
        if not watchdir['enabled']:
            # We shouldn't be updating because this watchdir is not enabled
            log.debug('Watchdir id %s is not enabled. Disabling it.', watchdir_id)
            self.disable_watchdir(watchdir_id)
            return

        if not os.path.isdir(watchdir['abspath']):
            log.warning('Invalid AutoAdd folder: %s', watchdir['abspath'])
            self.disable_watchdir(watchdir_id)
            return

        # Generate options dict for watchdir
        options = {}
        if 'stop_at_ratio_toggle' in watchdir:
            watchdir['stop_ratio_toggle'] = watchdir['stop_at_ratio_toggle']
        # We default to True when reading _toggle values, so a config
        # without them is valid, and applies all its settings.
        for option, value in watchdir.items():
            if OPTIONS_AVAILABLE.get(option):
                if watchdir.get(option + '_toggle', True) or option in [
                    'owner',
                    'seed_mode',
                ]:
                    options[option] = value

        # Check for .magnet files containing multiple magnet links and
        # create a new .magnet file for each of them.
        for filename in os.listdir(watchdir['abspath']):
            try:
                filepath = os.path.join(watchdir['abspath'], filename)
            except UnicodeDecodeError as ex:
                log.error(
                    'Unable to auto add torrent due to improper filename encoding: %s',
                    ex,
                )
                continue
            if os.path.isdir(filepath):
                # Skip directories
                continue
            elif os.path.splitext(filename)[1] == '.magnet' and self.split_magnets(
                filepath
            ):
                os.remove(filepath)

        for filename in os.listdir(watchdir['abspath']):
            try:
                filepath = os.path.join(watchdir['abspath'], filename)
            except UnicodeDecodeError as ex:
                log.error(
                    'Unable to auto add torrent due to improper filename encoding: %s',
                    ex,
                )
                continue

            if os.path.isdir(filepath):
                # Skip directories
                continue

            ext = os.path.splitext(filename)[1].lower()
            magnet = ext == '.magnet'
            if not magnet and not ext == '.torrent':
                log.debug('File checked for auto-loading is invalid: %s', filename)
                continue

            try:
                filedump = self.load_torrent(filepath, magnet)
            except (OSError, EOFError, InvalidTorrentError) as ex:
                # If torrent is invalid, keep track of it so can try again on the next pass.
                # This catches torrent files that may not be fully saved to disk at load time.
                log.debug('Torrent is invalid: %s', ex)
                if filename in self.invalid_torrents:
                    self.invalid_torrents[filename] += 1
                    if self.invalid_torrents[filename] >= MAX_NUM_ATTEMPTS:
                        log.warning(
                            'Maximum attempts reached while trying to add the '
                            'torrent file with the path %s',
                            filepath,
                        )
                        os.rename(filepath, filepath + '.invalid')
                        del self.invalid_torrents[filename]
                else:
                    self.invalid_torrents[filename] = 1
                continue

            def on_torrent_added(torrent_id, filename, filepath):
                if 'Label' in component.get('CorePluginManager').get_enabled_plugins():
                    if watchdir.get('label_toggle', True) and watchdir.get('label'):
                        label = component.get('CorePlugin.Label')
                        if not watchdir['label'] in label.get_labels():
                            label.add(watchdir['label'])
                        try:
                            label.set_torrent(torrent_id, watchdir['label'])
                        except Exception as ex:
                            log.error('Unable to set label: %s', ex)

                if (
                    watchdir.get('queue_to_top_toggle', True)
                    and 'queue_to_top' in watchdir
                ):
                    if watchdir['queue_to_top']:
                        component.get('TorrentManager').queue_top(torrent_id)
                    else:
                        component.get('TorrentManager').queue_bottom(torrent_id)

                # Rename, copy or delete the torrent once added to deluge.
                if watchdir.get('append_extension_toggle'):
                    if not watchdir.get('append_extension'):
                        watchdir['append_extension'] = '.added'
                    os.rename(filepath, filepath + watchdir['append_extension'])
                elif watchdir.get('copy_torrent_toggle'):
                    copy_torrent_path = watchdir['copy_torrent']
                    copy_torrent_file = os.path.join(copy_torrent_path, filename)
                    log.debug(
                        'Moving added torrent file "%s" to "%s"',
                        os.path.basename(filepath),
                        copy_torrent_path,
                    )
                    shutil.move(filepath, copy_torrent_file)
                else:
                    os.remove(filepath)

            def fail_torrent_add(err_msg, filepath, magnet):
                if isinstance(err_msg, Failure):
                    err_msg = err_msg.getErrorMessage()

                # torrent handle is invalid and so is the magnet link
                log.error(
                    'Cannot Autoadd %s: %s: %s',
                    'magnet' if magnet else 'torrent file',
                    filepath,
                    err_msg,
                )
                os.rename(filepath, filepath + '.invalid')

            try:
                # The torrent looks good, so lets add it to the session.
                if magnet:
                    d = maybeDeferred(
                        component.get('Core').add_torrent_magnet,
                        filedump.strip(),
                        options,
                    )
                else:
                    d = component.get('Core').add_torrent_file_async(
                        filename, b64encode(filedump), options
                    )
                d.addCallback(on_torrent_added, filename, filepath)
                d.addErrback(fail_torrent_add, filepath, magnet)
            except AddTorrentError as ex:
                fail_torrent_add(str(ex), filepath, magnet)

    def on_update_watchdir_error(self, failure, watchdir_id):
        """Disables any watch folders with un-handled exceptions."""
        self.disable_watchdir(watchdir_id)
        log.error(
            'Disabling "%s", error during update: %s',
            self.watchdirs[watchdir_id]['path'],
            failure,
        )

    @export
    def enable_watchdir(self, watchdir_id):
        w_id = str(watchdir_id)
        # Enable the looping call
        if w_id not in self.update_timers or not self.update_timers[w_id].running:
            self.update_timers[w_id] = LoopingCall(self.update_watchdir, w_id)
            self.update_timers[w_id].start(5).addErrback(
                self.on_update_watchdir_error, w_id
            )
        # Update the config
        if not self.watchdirs[w_id]['enabled']:
            self.watchdirs[w_id]['enabled'] = True
            self.config.save()
            component.get('EventManager').emit(AutoaddOptionsChangedEvent())

    @export
    def disable_watchdir(self, watchdir_id):
        w_id = str(watchdir_id)
        # Disable the looping call
        if w_id in self.update_timers:
            if self.update_timers[w_id].running:
                self.update_timers[w_id].stop()
            del self.update_timers[w_id]
        # Update the config
        if self.watchdirs[w_id]['enabled']:
            self.watchdirs[w_id]['enabled'] = False
            self.config.save()
            component.get('EventManager').emit(AutoaddOptionsChangedEvent())

    @export
    def set_config(self, config):
        """Sets the config dictionary."""
        config = self._make_unicode(config)
        for key in config:
            self.config[key] = config[key]
        self.config.save()
        component.get('EventManager').emit(AutoaddOptionsChangedEvent())

    @export
    def get_config(self):
        """Returns the config dictionary."""
        return self.config.config

    @export
    def get_watchdirs(self):
        session_user = self.rpcserver.get_session_user()
        session_auth_level = self.rpcserver.get_session_auth_level()
        if session_auth_level == AUTH_LEVEL_ADMIN:
            log.debug(
                'Current logged in user %s is an ADMIN, send all ' 'watchdirs',
                session_user,
            )
            return self.watchdirs

        watchdirs = {}
        for watchdir_id, watchdir in self.watchdirs.items():
            if watchdir.get('owner', 'localclient') == session_user:
                watchdirs[watchdir_id] = watchdir

        log.debug(
            'Current logged in user %s is not an ADMIN, send only '
            'their watchdirs: %s',
            session_user,
            list(watchdirs),
        )
        return watchdirs

    def _make_unicode(self, options):
        opts = {}
        for key in options:
            if isinstance(options[key], bytes):
                options[key] = options[key].decode('utf8')
            opts[key] = options[key]
        return opts

    @export
    def add(self, options=None):
        """Add a watch folder."""
        if options is None:
            options = {}
        options = self._make_unicode(options)
        abswatchdir = os.path.abspath(options['path'])
        check_input(os.path.isdir(abswatchdir), _('Path does not exist.'))
        check_input(
            os.access(abswatchdir, os.R_OK | os.W_OK),
            'You must have read and write access to watch folder.',
        )
        if abswatchdir in [wd['abspath'] for wd in self.watchdirs.values()]:
            raise Exception('Path is already being watched.')
        options.setdefault('enabled', False)
        options['abspath'] = abswatchdir
        watchdir_id = self.config['next_id']
        self.watchdirs[str(watchdir_id)] = options
        if options.get('enabled'):
            self.enable_watchdir(watchdir_id)
        self.config['next_id'] = watchdir_id + 1
        self.config.save()
        component.get('EventManager').emit(AutoaddOptionsChangedEvent())
        return watchdir_id

    @export
    def remove(self, watchdir_id):
        """Remove a watch folder."""
        watchdir_id = str(watchdir_id)
        check_input(
            watchdir_id in self.watchdirs, 'Unknown Watchdir: %s' % self.watchdirs
        )
        if self.watchdirs[watchdir_id]['enabled']:
            self.disable_watchdir(watchdir_id)
        del self.watchdirs[watchdir_id]
        self.config.save()
        component.get('EventManager').emit(AutoaddOptionsChangedEvent())

    def __migrate_config_1_to_2(self, config):
        for watchdir_id in config['watchdirs']:
            config['watchdirs'][watchdir_id]['owner'] = 'localclient'
        return config

    def __on_pre_torrent_removed(self, torrent_id):
        try:
            torrent = component.get('TorrentManager')[torrent_id]
        except KeyError:
            log.warning(
                'Unable to remove torrent file for torrent id %s. It'
                'was already deleted from the TorrentManager',
                torrent_id,
            )
            return
        torrent_fname = torrent.filename
        for watchdir in self.watchdirs.values():
            if not watchdir.get('copy_torrent_toggle', False):
                # This watchlist does copy torrents
                continue
            elif not watchdir.get('delete_copy_torrent_toggle', False):
                # This watchlist is not set to delete finished torrents
                continue
            copy_torrent_path = watchdir['copy_torrent']
            torrent_fname_path = os.path.join(copy_torrent_path, torrent_fname)
            if os.path.isfile(torrent_fname_path):
                try:
                    os.remove(torrent_fname_path)
                    log.info(
                        'Removed torrent file "%s" from "%s"',
                        torrent_fname,
                        copy_torrent_path,
                    )
                    break
                except OSError as ex:
                    log.info(
                        'Failed to removed torrent file "%s" from "%s": %s',
                        torrent_fname,
                        copy_torrent_path,
                        ex,
                    )

    @export
    def is_admin_level(self):
        return self.rpcserver.get_session_auth_level() == deluge.common.AUTH_LEVEL_ADMIN

    @export
    def get_auth_user(self):
        return self.rpcserver.get_session_user()



#
# Copyright (C) 2009 Ian Martin <ianmartin@cantab.net>
# Copyright (C) 2008 Damien Churchill <damoxc@gmail.com>
# Copyright (C) 2008 Martijn Voncken <mvoncken@gmail.com>
# Copyright (C) 2007 Marcos Mobley <markybob@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging
import time

from twisted.internet.task import LoopingCall

from deluge import component, configmanager
from deluge.core.rpcserver import export
from deluge.plugins.pluginbase import CorePluginBase

DEFAULT_PREFS = {
    'test': 'NiNiNi',
    'update_interval': 1,  # 2 seconds.
    'length': 150,  # 2 seconds * 150 --> 5 minutes.
}

DEFAULT_TOTALS = {
    'total_upload': 0,
    'total_download': 0,
    'total_payload_upload': 0,
    'total_payload_download': 0,
    'stats': {},
}

log = logging.getLogger(__name__)


def get_key(config, key):
    try:
        return config[key]
    except KeyError:
        return None


def mean(items):
    try:
        return sum(items) // len(items)
    except Exception:
        return 0


class Core(CorePluginBase):
    totals = {}  # class var to catch only updating this once per session in enable.

    def enable(self):
        log.debug('Stats plugin enabled')
        self.core = component.get('Core')
        self.stats = {}
        self.count = {}
        self.intervals = [1, 5, 30, 300]

        self.last_update = {}
        t = time.time()
        for i in self.intervals:
            self.stats[i] = {}
            self.last_update[i] = t
            self.count[i] = 0

        self.config = configmanager.ConfigManager('stats.conf', DEFAULT_PREFS)
        self.saved_stats = configmanager.ConfigManager('stats.totals', DEFAULT_TOTALS)
        if self.totals == {}:
            self.totals.update(self.saved_stats.config)

        self.length = self.config['length']

        # self.stats = get_key(self.saved_stats, 'stats') or {}

        # keys needed from core.get_session_status
        self.stat_keys = [
            'upload_rate',
            'download_rate',
            'dht.dht_nodes',
            'dht.dht_node_cache',
            'dht.dht_torrents',
            'peer.num_peers_connected',
            'peer.num_peers_half_open',
        ]
        # collected statistics and functions to get them
        self.stat_getters = {
            'upload_rate': lambda s: s['upload_rate'],
            'download_rate': lambda s: s['download_rate'],
            'dht_nodes': lambda s: s['dht.dht_nodes'],
            'dht_cache_nodes': lambda s: s['dht.dht_node_cache'],
            'dht_torrents': lambda s: s['dht.dht_torrents'],
            'num_peers': lambda s: s['peer.num_peers_connected'],
            'num_connections': lambda s: s['peer.num_peers_connected']
            + s['peer.num_peers_half_open'],
        }

        # initialize stats object
        for key in self.stat_getters.keys():
            for i in self.intervals:
                if key not in self.stats[i]:
                    self.stats[i][key] = []

        self.update_stats()

        self.update_timer = LoopingCall(self.update_stats)
        self.update_timer.start(self.config['update_interval'])

        self.save_timer = LoopingCall(self.save_stats)
        self.save_timer.start(60)

    def disable(self):
        self.update_timer.stop() if self.update_timer.running else None
        self.save_timer.stop() if self.save_timer.running else None
        self.save_stats()

    def update_stats(self):
        # Get all possible stats!
        stats = {}
        raw_stats = self.core.get_session_status(self.stat_keys)
        for name, fn in self.stat_getters.items():
            stats[name] = fn(raw_stats)

        # status = self.core.session.status()
        # for stat in dir(status):
        #     if not stat.startswith('_') and stat not in stats:
        #         stats[stat] = getattr(status, stat, None)

        update_time = time.time()
        self.last_update[1] = update_time

        # extract the ones we are interested in
        # adding them to the 1s array
        for stat, stat_list in self.stats[1].items():
            if stat in stats:
                stat_list.insert(0, int(stats[stat]))
            else:
                stat_list.insert(0, 0)
            if len(stat_list) > self.length:
                stat_list.pop()

        def update_interval(interval, base, multiplier):
            self.count[interval] = self.count[interval] + 1
            if self.count[interval] >= interval:
                self.last_update[interval] = update_time
                self.count[interval] = 0
                current_stats = self.stats[interval]
                for stat, stat_list in self.stats[base].items():
                    try:
                        avg = mean(stat_list[0:multiplier])
                    except ValueError:
                        avg = 0
                    current_stats[stat].insert(0, avg)
                    if len(current_stats[stat]) > self.length:
                        current_stats[stat].pop()

        update_interval(5, 1, 5)
        update_interval(30, 5, 6)
        update_interval(300, 30, 10)

    def save_stats(self):
        self.saved_stats['stats'] = self.stats
        self.saved_stats.config.update(self.get_totals())
        self.saved_stats.save()

    # export:
    @export
    def get_stats(self, keys, interval):
        if interval not in self.intervals:
            return None

        stats_dict = {}
        for key in keys:
            if key in self.stats[interval]:
                stats_dict[key] = self.stats[interval][key]

        stats_dict['_last_update'] = self.last_update[interval]
        stats_dict['_length'] = self.config['length']
        stats_dict['_update_interval'] = interval
        return stats_dict

    @export
    def get_totals(self):
        result = {}
        session_totals = self.get_session_totals()
        for key in session_totals:
            result[key] = self.totals[key] + session_totals[key]
        return result

    @export
    def get_session_totals(self):
        return self.core.get_session_status(
            [
                'total_upload',
                'total_download',
                'total_payload_upload',
                'total_payload_download',
            ]
        )

    @export
    def set_config(self, config):
        """Sets the config dictionary."""
        for key in config:
            self.config[key] = config[key]
        self.config.save()

    @export
    def get_config(self):
        """Returns the config dictionary."""
        return self.config.config

    @export
    def get_intervals(self):
        """Returns the available resolutions."""
        return self.intervals



#
# Copyright (C) 2009 Ian Martin <ianmartin@cantab.net>
# Copyright (C) 2008 Damien Churchill <damoxc@gmail.com>
# Copyright (C) 2008 Martijn Voncken <mvoncken@gmail.com>
# Copyright (C) 2007 Marcos Mobley <markybob@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

"""
port of old plugin by markybob.
"""

import logging
import math
import time

import gi

gi.require_foreign('cairo')

import cairo  # isort:skip (gi checks required before import).

log = logging.getLogger(__name__)

black = (0, 0, 0)
gray = (0.75, 0.75, 0.75)
white = (1.0, 1.0, 1.0)
darkred = (0.65, 0, 0)
red = (1.0, 0, 0)
green = (0, 1.0, 0)
blue = (0, 0, 1.0)
orange = (1.0, 0.74, 0)


def default_formatter(value):
    return str(value)


def size_formatter_scale(value):
    scale = 1.0
    for i in range(0, 3):
        scale = scale * 1024.0
        if value // scale < 1024:
            return scale


def change_opacity(color, opactiy):
    """A method to assist in changing the opactiy of a color inorder to draw the
    fills.
    """
    color = list(color)
    if len(color) == 4:
        color[3] = opactiy
    else:
        color.append(opactiy)
    return tuple(color)


class Graph:
    def __init__(self):
        self.width = 100
        self.height = 100
        self.length = 150
        self.stat_info = {}
        self.line_size = 2
        self.dash_length = [10]
        self.mean_selected = True
        self.legend_selected = True
        self.max_selected = True
        self.black = (0, 0, 0)
        self.interval = 2  # 2 secs
        self.text_bg = (255, 255, 255, 128)  # prototyping
        self.set_left_axis()

    def set_left_axis(self, **kargs):
        self.left_axis = kargs

    def add_stat(self, stat, label='', axis='left', line=True, fill=True, color=None):
        self.stat_info[stat] = {
            'axis': axis,
            'label': label,
            'line': line,
            'fill': fill,
            'color': color,
        }

    def set_stats(self, stats):
        self.last_update = stats['_last_update']
        del stats['_last_update']
        self.length = stats['_length']
        del stats['_length']
        self.interval = stats['_update_interval']
        del stats['_update_interval']
        self.stats = stats
        return

    # def set_config(self, config):
    #      self.length = config["length"]
    #      self.interval = config["update_interval"]

    def set_interval(self, interval):
        self.interval = interval

    def draw_to_context(self, context, width, height):
        self.ctx = context
        self.width, self.height = width, height
        self.draw_rect(white, 0, 0, self.width, self.height)
        self.draw_graph()
        return self.ctx

    def draw(self, width, height):
        surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, width, height)
        ctx = cairo.Context(surface)
        self.draw_to_context(ctx, width, height)
        return surface

    def draw_x_axis(self, bounds):
        (left, top, right, bottom) = bounds
        duration = self.length * self.interval
        start = self.last_update - duration
        ratio = (right - left) / duration

        if duration < 1800 * 10:
            # try rounding to nearest 1min, 5mins, 10mins, 30mins
            for step in [60, 300, 600, 1800]:
                if duration // step < 10:
                    x_step = step
                    break
        else:
            # If there wasnt anything useful find a nice fitting hourly divisor
            x_step = ((duration // 5) // 3600) * 3600

        # this doesnt allow for dst and timezones...
        seconds_to_step = math.ceil(start / x_step) * x_step - start

        for i in range(0, duration // x_step + 1):
            text = time.strftime(
                '%H:%M', time.localtime(start + seconds_to_step + i * x_step)
            )
            # + 0.5 to allign x to nearest pixel
            x = int(ratio * (seconds_to_step + i * x_step) + left) + 0.5
            self.draw_x_text(text, x, bottom)
            self.draw_dotted_line(gray, x, top - 0.5, x, bottom + 0.5)

        self.draw_line(gray, left, bottom + 0.5, right, bottom + 0.5)

    def draw_graph(self):
        font_extents = self.ctx.font_extents()
        x_axis_space = font_extents[2] + 2 + self.line_size / 2
        plot_height = self.height - x_axis_space
        # lets say we need 2n-1*font height pixels to plot the y ticks
        tick_limit = plot_height / font_extents[3]

        max_value = 0
        for stat in self.stat_info:
            if self.stat_info[stat]['axis'] == 'left':
                try:
                    l_max = max(self.stats[stat])
                except ValueError:
                    l_max = 0
                if l_max > max_value:
                    max_value = l_max
        if max_value < self.left_axis['min']:
            max_value = self.left_axis['min']

        y_ticks = self.intervalise(max_value, tick_limit)
        max_value = y_ticks[-1]
        # find the width of the y_ticks
        y_tick_text = [self.left_axis['formatter'](tick) for tick in y_ticks]

        def space_required(text):
            te = self.ctx.text_extents(text)
            return math.ceil(te[4] - te[0])

        y_tick_width = max(space_required(text) for text in y_tick_text)

        top = font_extents[2] / 2
        # bounds(left, top, right, bottom)
        bounds = (y_tick_width + 4, top + 2, self.width, self.height - x_axis_space)

        self.draw_x_axis(bounds)
        self.draw_left_axis(bounds, y_ticks, y_tick_text)

    def intervalise(self, x, limit=None):
        """Given a value x create an array of tick points to got with the graph
        The number of ticks returned can be constrained by limit, minimum of 3
        """
        # Limit is the number of ticks which is 1 + the number of steps as we
        # count the 0 tick in limit
        if limit is not None:
            if limit < 3:
                limit = 2
            else:
                limit = limit - 1
        scale = 1
        if 'formatter_scale' in self.left_axis:
            scale = self.left_axis['formatter_scale'](x)
            x = x / scale

        # Find the largest power of 10 less than x
        comm_log = math.log10(x)
        intbit = math.floor(comm_log)

        interval = math.pow(10, intbit)
        steps = int(math.ceil(x / interval))

        if steps <= 1 and (limit is None or limit >= 10 * steps):
            interval = interval * 0.1
            steps = steps * 10
        elif steps <= 2 and (limit is None or limit >= 5 * steps):
            interval = interval * 0.2
            steps = steps * 5
        elif steps <= 5 and (limit is None or limit >= 2 * steps):
            interval = interval * 0.5
            steps = steps * 2

        if limit is not None and steps > limit:
            multi = steps / limit
            if multi > 2:
                interval = interval * 5
            else:
                interval = interval * 2

        intervals = [
            i * interval * scale for i in range(1 + int(math.ceil(x / interval)))
        ]
        return intervals

    def draw_left_axis(self, bounds, y_ticks, y_tick_text):
        (left, top, right, bottom) = bounds
        stats = {}
        for stat in self.stat_info:
            if self.stat_info[stat]['axis'] == 'left':
                stats[stat] = self.stat_info[stat]
                stats[stat]['values'] = self.stats[stat]
                stats[stat]['fill_color'] = change_opacity(stats[stat]['color'], 0.5)
                stats[stat]['color'] = change_opacity(stats[stat]['color'], 0.8)

        height = bottom - top
        max_value = y_ticks[-1]
        ratio = height / max_value

        for i, y_val in enumerate(y_ticks):
            y = int(bottom - y_val * ratio) - 0.5
            if i != 0:
                self.draw_dotted_line(gray, left, y, right, y)
            self.draw_y_text(y_tick_text[i], left, y)
        self.draw_line(gray, left, top, left, bottom)

        for stat, info in stats.items():
            if len(info['values']) > 0:
                self.draw_value_poly(info['values'], info['color'], max_value, bounds)
                self.draw_value_poly(
                    info['values'], info['fill_color'], max_value, bounds, info['fill']
                )

    def draw_legend(self):
        pass

    def trace_path(self, values, max_value, bounds):
        (left, top, right, bottom) = bounds
        ratio = (bottom - top) / max_value
        line_width = self.line_size

        self.ctx.set_line_width(line_width)
        self.ctx.move_to(right, bottom)

        self.ctx.line_to(right, int(bottom - values[0] * ratio))

        x = right
        step = (right - left) / (self.length - 1)
        for i, value in enumerate(values):
            if i == self.length - 1:
                x = left

            self.ctx.line_to(x, int(bottom - value * ratio))
            x -= step

        self.ctx.line_to(int(right - (len(values) - 1) * step), bottom)
        self.ctx.close_path()

    def draw_value_poly(self, values, color, max_value, bounds, fill=False):
        self.trace_path(values, max_value, bounds)
        self.ctx.set_source_rgba(*color)

        if fill:
            self.ctx.fill()
        else:
            self.ctx.stroke()

    def draw_x_text(self, text, x, y):
        """Draws text below and horizontally centered about x,y"""
        fe = self.ctx.font_extents()
        te = self.ctx.text_extents(text)
        height = fe[2]
        x_bearing = te[0]
        width = te[2]
        self.ctx.move_to(int(x - width / 2 + x_bearing), int(y + height))
        self.ctx.set_source_rgba(*self.black)
        self.ctx.show_text(text)

    def draw_y_text(self, text, x, y):
        """Draws text left of and vertically centered about x,y"""
        fe = self.ctx.font_extents()
        te = self.ctx.text_extents(text)
        descent = fe[1]
        ascent = fe[0]
        x_bearing = te[0]
        width = te[4]
        self.ctx.move_to(
            int(x - width - x_bearing - 2), int(y + (ascent - descent) / 2)
        )
        self.ctx.set_source_rgba(*self.black)
        self.ctx.show_text(text)

    def draw_rect(self, color, x, y, height, width):
        self.ctx.set_source_rgba(*color)
        self.ctx.rectangle(x, y, height, width)
        self.ctx.fill()

    def draw_line(self, color, x1, y1, x2, y2):
        self.ctx.set_source_rgba(*color)
        self.ctx.set_line_width(1)
        self.ctx.move_to(x1, y1)
        self.ctx.line_to(x2, y2)
        self.ctx.stroke()

    def draw_dotted_line(self, color, x1, y1, x2, y2):
        self.ctx.set_source_rgba(*color)
        self.ctx.set_line_width(1)
        dash, offset = self.ctx.get_dash()
        self.ctx.set_dash(self.dash_length, 0)
        self.ctx.move_to(x1, y1)
        self.ctx.line_to(x2, y2)
        self.ctx.stroke()
        self.ctx.set_dash(dash, offset)



#
# Copyright (C) 2008-2009 Ido Abramovich <ido.deluge@gmail.com>
# Copyright (C) 2009 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging

import deluge.component as component
from deluge.ui.client import client

from . import BaseCommand

log = logging.getLogger(__name__)


class Command(BaseCommand):
    """Connect to a new deluge server"""

    usage = _('Usage: connect <host[:port]> [<username>] [<password>]')

    def add_arguments(self, parser):
        parser.add_argument(
            'host', help=_('Daemon host and port'), metavar='<host[:port]>'
        )
        parser.add_argument(
            'username', help=_('Username'), metavar='<username>', nargs='?', default=''
        )
        parser.add_argument(
            'password', help=_('Password'), metavar='<password>', nargs='?', default=''
        )

    def add_parser(self, subparsers):
        parser = subparsers.add_parser(
            self.name, help=self.__doc__, description=self.__doc__, prog='connect'
        )
        self.add_arguments(parser)

    def handle(self, options):
        self.console = component.get('ConsoleUI')

        host = options.host
        try:
            host, port = host.split(':')
            port = int(port)
        except ValueError:
            port = 58846

        def do_connect():
            d = client.connect(host, port, options.username, options.password)

            def on_connect(result):
                if self.console.interactive:
                    self.console.write(f'{{!success!}}Connected to {host}:{port}!')
                return component.start()

            def on_connect_fail(result):
                self.console.write(
                    f'{{!error!}}Failed to connect to {host}:{port} with reason: {result.value.message}'
                )
                return result

            d.addCallbacks(on_connect, on_connect_fail)
            return d

        if client.connected():

            def on_disconnect(result):
                if self.console.statusbars:
                    self.console.statusbars.update_statusbars()
                return do_connect()

            return client.disconnect().addCallback(on_disconnect)
        else:
            return do_connect()



#!/usr/bin/env python
# Written by Martin v. Lwis <loewis@informatik.hu-berlin.de>
# Plural forms support added by alexander smishlajev <alex@tycobka.lv>
"""
Generate binary message catalog from textual translation description.

This program converts a textual Uniforum-style message catalog (.po file) into
a binary GNU catalog (.mo file).  This is essentially the same function as the
GNU msgfmt program, however, it is a simpler implementation.

Usage: msgfmt.py [OPTIONS] filename.po

Options:
    -o file
    --output-file=file
        Specify the output file to write to.  If omitted, output will go to a
        file named filename.mo (based off the input file name).

    -h
    --help
        Print this message and exit.

    -V
    --version
        Display version information and exit.
"""
import array
import ast
import getopt
import os
import struct
import sys

__version__ = '1.2'

MESSAGES = {}


def usage(ecode, msg=''):
    """
    Print usage and msg and exit with given code.
    """
    print(__doc__, file=sys.stderr)
    if msg:
        print(msg, file=sys.stderr)
    sys.exit(ecode)


def add(msgid, transtr, fuzzy):
    """
    Add a non-fuzzy translation to the dictionary.
    """
    if not fuzzy and transtr and not transtr.startswith('\x00'):
        MESSAGES[msgid] = transtr


def generate():
    """
    Return the generated output.
    """
    # the keys are sorted in the .mo file
    keys = sorted(MESSAGES)
    offsets = []
    ids = strs = ''
    for _id in keys:
        # For each string, we need size and file offset when encoded. Each string is NUL
        # terminated; the NUL does not count into the size.
        offsets.append(
            (
                len(ids.encode('utf8')),
                len(_id.encode('utf8')),
                len(strs.encode('utf8')),
                len(MESSAGES[_id].encode('utf8')),
            )
        )
        ids += _id + '\x00'
        strs += MESSAGES[_id] + '\x00'

    # The header is 7 32-bit unsigned integers.  We don't use hash tables, so
    # the keys start right after the index tables.
    # translated string.
    keystart = 7 * 4 + 16 * len(keys)
    # and the values start after the keys
    valuestart = keystart + len(ids)
    koffsets = []
    voffsets = []
    # The string table first has the list of keys, then the list of values.
    # Each entry has first the size of the string, then the file offset.
    for o1, l1, o2, l2 in offsets:
        koffsets += [l1, o1 + keystart]
        voffsets += [l2, o2 + valuestart]
    offsets = koffsets + voffsets
    output = struct.pack(
        'Iiiiiii',
        0x950412DE,  # Magic
        0,  # Version
        len(keys),  # # of entries
        7 * 4,  # start of key index
        7 * 4 + len(keys) * 8,  # start of value index
        0,
        0,
    )  # size and offset of hash table
    output += array.array('i', offsets).tobytes()
    output += ids.encode('utf8')
    output += strs.encode('utf8')
    return output


def make(filename, outfile):
    section_id = 1
    section_str = 2
    global MESSAGES
    MESSAGES = {}

    # Compute .mo name from .po name and arguments
    if filename.endswith('.po'):
        infile = filename
    else:
        infile = filename + '.po'
    if outfile is None:
        outfile = os.path.splitext(infile)[0] + '.mo'

    try:
        with open(infile, encoding='utf8') as _file:
            lines = _file.readlines()
    except OSError as msg:
        print(msg, file=sys.stderr)
        sys.exit(1)

    section = None
    fuzzy = 0

    # Parse the catalog
    msgid = msgstr = ''
    lno = 0
    for line in lines:
        lno += 1
        # If we get a comment line after a msgstr, this is a new entry
        if line[0] == '#' and section == section_str:
            add(msgid, msgstr, fuzzy)
            section = None
            fuzzy = 0
        # Record a fuzzy mark
        if line[:2] == '#,' and (line.find('fuzzy') >= 0):
            fuzzy = 1
        # Skip comments
        if line[0] == '#':
            continue
        # Start of msgid_plural section, separate from singular form with \0
        if line.startswith('msgid_plural'):
            msgid += '\x00'
            line = line[12:]
        # Now we are in a msgid section, output previous section
        elif line.startswith('msgid'):
            if section == section_str:
                add(msgid, msgstr, fuzzy)
            section = section_id
            line = line[5:]
            msgid = msgstr = ''
        # Now we are in a msgstr section
        elif line.startswith('msgstr'):
            section = section_str
            line = line[6:]
            # Check for plural forms
            if line.startswith('['):
                # Separate plural forms with \0
                if not line.startswith('[0]'):
                    msgstr += '\x00'
                # Ignore the index - must come in sequence
                line = line[line.index(']') + 1 :]
        # Skip empty lines
        line = line.strip()
        if not line:
            continue
        line = ast.literal_eval(line)
        if section == section_id:
            msgid += line
        elif section == section_str:
            msgstr += line
        else:
            print('Syntax error on %s:%d' % (infile, lno), 'before:', file=sys.stderr)
            print(line, file=sys.stderr)
            sys.exit(1)
    # Add last entry
    if section == section_str:
        add(msgid, msgstr, fuzzy)

    # Compute output
    output = generate()

    try:
        with open(outfile, 'wb') as _file:
            _file.write(output)
    except OSError as msg:
        print(msg, file=sys.stderr)


def main():
    try:
        opts, args = getopt.getopt(
            sys.argv[1:], 'hVo:', ['help', 'version', 'output-file=']
        )
    except getopt.error as msg:
        usage(1, msg)

    outfile = None
    # parse options
    for opt, arg in opts:
        if opt in ('-h', '--help'):
            usage(0)
        elif opt in ('-V', '--version'):
            print('msgfmt.py', __version__, file=sys.stderr)
            sys.exit(0)
        elif opt in ('-o', '--output-file'):
            outfile = arg
    # do it
    if not args:
        print('No input file given', file=sys.stderr)
        print('Try `msgfmt --help` for more information.', file=sys.stderr)
        return

    for filename in args:
        make(filename, outfile)


if __name__ == '__main__':
    main()



#
# Copyright (C) 2009-2010 Pedro Algarvio <pedro@algarvio.me>
#
# Basic plugin template created by:
# Copyright (C) 2008 Martijn Voncken <mvoncken@gmail.com>
# Copyright (C) 2007-2009 Andrew Resch <andrewresch@gmail.com>
# Copyright (C) 2009 Damien Churchill <damoxc@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging
from os.path import basename

from gi import require_version
from gi.repository import Gtk
from twisted.internet import defer

import deluge.common
import deluge.component as component
import deluge.configmanager
from deluge.plugins.pluginbase import Gtk3PluginBase
from deluge.ui.client import client

from .common import CustomNotifications, get_resource

# Relative imports

log = logging.getLogger(__name__)

try:
    import pygame

    SOUND_AVAILABLE = True
except ImportError:
    SOUND_AVAILABLE = False

try:
    require_version('Notify', '0.7')
    from gi.repository import GLib, Notify
except (ValueError, ImportError):
    POPUP_AVAILABLE = False
else:
    POPUP_AVAILABLE = not deluge.common.windows_check()


DEFAULT_PREFS = {
    # BLINK
    'blink_enabled': False,
    # FLASH
    'flash_enabled': False,
    # POPUP
    'popup_enabled': False,
    # SOUND
    'sound_enabled': False,
    'sound_path': '',
    'custom_sounds': {},
    # Subscriptions
    'subscriptions': {'popup': [], 'blink': [], 'sound': []},
}

RECIPIENT_FIELD, RECIPIENT_EDIT = list(range(2))
(
    SUB_EVENT,
    SUB_EVENT_DOC,
    SUB_NOT_EMAIL,
    SUB_NOT_POPUP,
    SUB_NOT_BLINK,
    SUB_NOT_SOUND,
) = list(range(6))
SND_EVENT, SND_EVENT_DOC, SND_NAME, SND_PATH = list(range(4))


class GtkUiNotifications(CustomNotifications):
    def __init__(self, plugin_name=None):
        CustomNotifications.__init__(self, plugin_name)

    def enable(self):
        CustomNotifications.enable(self)
        self.register_custom_blink_notification(
            'TorrentFinishedEvent', self._on_torrent_finished_event_blink
        )
        self.register_custom_sound_notification(
            'TorrentFinishedEvent', self._on_torrent_finished_event_sound
        )
        self.register_custom_popup_notification(
            'TorrentFinishedEvent', self._on_torrent_finished_event_popup
        )

    def disable(self):
        self.deregister_custom_blink_notification('TorrentFinishedEvent')
        self.deregister_custom_sound_notification('TorrentFinishedEvent')
        self.deregister_custom_popup_notification('TorrentFinishedEvent')
        CustomNotifications.disable(self)

    def register_custom_popup_notification(self, eventtype, handler):
        """This is used to register popup notifications for custom event types.

        :param event: the event name
        :param type: string
        :param handler: function, to be called when `:param:event` is emitted

        Your handler should return a tuple of (popup_title, popup_contents).
        """
        self._register_custom_provider('popup', eventtype, handler)

    def deregister_custom_popup_notification(self, eventtype):
        self._deregister_custom_provider('popup', eventtype)

    def register_custom_blink_notification(self, eventtype, handler):
        """This is used to register blink notifications for custom event types.

        :param event: str, the event name
        :param handler: function, to be called when `:param:event` is emitted

        Your handler should return `True` or `False` to blink or not the
        trayicon.
        """
        self._register_custom_provider('blink', eventtype, handler)

    def deregister_custom_blink_notification(self, eventtype):
        self._deregister_custom_provider('blink', eventtype)

    def register_custom_sound_notification(self, eventtype, handler):
        """This is used to register sound notifications for custom event types.

        :param event: the event name
        :type event: string
        :param handler: function to be called when `:param:event` is emitted

        Your handler should return either '' to use the sound defined on the
        notification preferences, the path to a sound file, which will then be
        played or None, where no sound will be played at all.
        """
        self._register_custom_provider('sound', eventtype, handler)

    def deregister_custom_sound_notification(self, eventtype):
        self._deregister_custom_provider('sound', eventtype)

    def handle_custom_popup_notification(self, result, eventtype):
        title, message = result
        return defer.maybeDeferred(self.__popup, title, message)

    def handle_custom_blink_notification(self, result, eventtype):
        if result:
            return defer.maybeDeferred(self.__blink)
        return defer.succeed(
            'Will not blink. The returned value from the custom '
            'handler was: %s' % result
        )

    def handle_custom_sound_notification(self, result, eventtype):
        if isinstance(result, ''.__class__):
            if not result and eventtype in self.config['custom_sounds']:
                return defer.maybeDeferred(
                    self.__play_sound, self.config['custom_sounds'][eventtype]
                )
            return defer.maybeDeferred(self.__play_sound, result)
        return defer.succeed(
            'Will not play sound. The returned value from the '
            'custom handler was: %s' % result
        )

    def __blink(self):
        self.systray.blink(True)
        return defer.succeed(_('Notification Blink shown'))

    def __popup(self, title='', message=''):
        if not self.config['popup_enabled']:
            return defer.succeed(_('Popup notification is not enabled.'))
        if not POPUP_AVAILABLE:
            err_msg = _('libnotify is not installed')
            log.warning(err_msg)
            return defer.fail(ImportError(err_msg))

        if Notify.init('Deluge'):
            self.note = Notify.Notification.new(title, message, 'deluge-panel')
            self.note.set_hint('desktop-entry', GLib.Variant.new_string('deluge'))
            if not self.note.show():
                err_msg = _('Failed to popup notification')
                log.warning(err_msg)
                return defer.fail(Exception(err_msg))
        return defer.succeed(_('Notification popup shown'))

    def __play_sound(self, sound_path=''):
        if not self.config['sound_enabled']:
            return defer.succeed(_('Sound notification not enabled'))
        if not SOUND_AVAILABLE:
            err_msg = _('pygame is not installed')
            log.warning(err_msg)
            return defer.fail(ImportError(err_msg))

        pygame.init()
        try:
            if not sound_path:
                sound_path = self.config['sound_path']
            alert_sound = pygame.mixer.music
            alert_sound.load(sound_path)
            alert_sound.play()
        except pygame.error as ex:
            err_msg = _('Sound notification failed %s') % ex
            log.warning(err_msg)
            return defer.fail(ex)
        else:
            msg = _('Sound notification Success')
            log.info(msg)
            return defer.succeed(msg)

    def _on_torrent_finished_event_blink(self, torrent_id):
        return True  # Yes, Blink

    def _on_torrent_finished_event_sound(self, torrent_id):
        # Since there's no custom sound hardcoded, just return ''
        return ''

    def _on_torrent_finished_event_popup(self, torrent_id):
        d = client.core.get_torrent_status(torrent_id, ['name', 'file_progress'])
        d.addCallback(self._on_torrent_finished_event_got_torrent_status)
        d.addErrback(self._on_torrent_finished_event_torrent_status_failure)
        return d

    def _on_torrent_finished_event_torrent_status_failure(self, failure):
        log.debug('Failed to get torrent status to be able to show the popup')

    def _on_torrent_finished_event_got_torrent_status(self, torrent_status):
        log.debug(
            'Handler for TorrentFinishedEvent GTKUI called. ' 'Got Torrent Status'
        )
        title = _('Finished Torrent')
        torrent_status['num_files'] = torrent_status['file_progress'].count(1.0)
        message = (
            _(
                'The torrent "%(name)s" including %(num_files)i file(s) '
                'has finished downloading.'
            )
            % torrent_status
        )
        return title, message


class GtkUI(Gtk3PluginBase, GtkUiNotifications):
    def __init__(self, plugin_name):
        Gtk3PluginBase.__init__(self, plugin_name)
        GtkUiNotifications.__init__(self)

    def enable(self):
        self.config = deluge.configmanager.ConfigManager(
            'notifications-gtk.conf', DEFAULT_PREFS
        )
        self.builder = Gtk.Builder()
        self.builder.add_from_file(get_resource('config.ui'))
        self.builder.get_object('smtp_port').set_value(25)
        self.prefs = self.builder.get_object('prefs_box')
        self.prefs.show_all()

        self.build_recipients_model_populate_treeview()
        self.build_sounds_model_populate_treeview()
        self.build_notifications_model_populate_treeview()

        client.notifications.get_handled_events().addCallback(
            self.popuplate_what_needs_handled_events
        )

        self.builder.connect_signals(
            {
                'on_add_button_clicked': (
                    self.on_add_button_clicked,
                    self.recipients_treeview,
                ),
                'on_delete_button_clicked': (
                    self.on_delete_button_clicked,
                    self.recipients_treeview,
                ),
                'on_enabled_toggled': self.on_enabled_toggled,
                'on_sound_enabled_toggled': self.on_sound_enabled_toggled,
                'on_sounds_edit_button_clicked': self.on_sounds_edit_button_clicked,
                'on_sounds_revert_button_clicked': self.on_sounds_revert_button_clicked,
                'on_sound_path_update_preview': self.on_sound_path_update_preview,
            }
        )

        component.get('Preferences').add_page(_('Notifications'), self.prefs)

        component.get('PluginManager').register_hook(
            'on_apply_prefs', self.on_apply_prefs
        )
        component.get('PluginManager').register_hook(
            'on_show_prefs', self.on_show_prefs
        )

        if not POPUP_AVAILABLE:
            self.builder.get_object('popup_enabled').set_property('sensitive', False)
        if not SOUND_AVAILABLE:
            # for widget_name in ('sound_enabled', 'sound_path', 'sounds_page', 'sounds_page_label'):
            #    self.builder.get_object(widget_name).set_property('sensitive', False)
            self.builder.get_object('sound_enabled').set_property('sensitive', False)
            self.builder.get_object('sound_path').set_property('sensitive', False)
            self.builder.get_object('sounds_page').set_property('sensitive', False)
            self.builder.get_object('sounds_page_label').set_property(
                'sensitive', False
            )

        self.systray = component.get('SystemTray')
        if not hasattr(self.systray, 'tray'):
            # Tray is not beeing used
            self.builder.get_object('blink_enabled').set_property('sensitive', False)

        GtkUiNotifications.enable(self)

    def disable(self):
        GtkUiNotifications.disable(self)
        component.get('Preferences').remove_page(_('Notifications'))
        component.get('PluginManager').deregister_hook(
            'on_apply_prefs', self.on_apply_prefs
        )
        component.get('PluginManager').deregister_hook(
            'on_show_prefs', self.on_show_prefs
        )

    def build_recipients_model_populate_treeview(self):
        # SMTP Recipients treeview/model
        self.recipients_treeview = self.builder.get_object('smtp_recipients')
        treeview_selection = self.recipients_treeview.get_selection()
        treeview_selection.connect(
            'changed', self.on_recipients_treeview_selection_changed
        )
        self.recipients_model = Gtk.ListStore(str, bool)

        renderer = Gtk.CellRendererText()
        renderer.connect('edited', self.on_cell_edited, self.recipients_model)
        renderer.recipient = RECIPIENT_FIELD
        column = Gtk.TreeViewColumn(
            'Recipients', renderer, text=RECIPIENT_FIELD, editable=RECIPIENT_EDIT
        )
        column.set_expand(True)
        self.recipients_treeview.append_column(column)
        self.recipients_treeview.set_model(self.recipients_model)

    def build_sounds_model_populate_treeview(self):
        # Sound customisation treeview/model
        self.sounds_treeview = self.builder.get_object('sounds_treeview')
        sounds_selection = self.sounds_treeview.get_selection()
        sounds_selection.connect('changed', self.on_sounds_treeview_selection_changed)

        self.sounds_treeview.set_tooltip_column(SND_EVENT_DOC)
        self.sounds_model = Gtk.ListStore(str, str, str, str)

        renderer = Gtk.CellRendererText()
        renderer.event = SND_EVENT
        column = Gtk.TreeViewColumn('Event', renderer, text=SND_EVENT)
        column.set_expand(True)
        self.sounds_treeview.append_column(column)

        renderer = Gtk.CellRendererText()
        renderer.event_doc = SND_EVENT_DOC
        column = Gtk.TreeViewColumn('Doc', renderer, text=SND_EVENT_DOC)
        column.set_property('visible', False)
        self.sounds_treeview.append_column(column)

        renderer = Gtk.CellRendererText()
        renderer.sound_name = SND_NAME
        column = Gtk.TreeViewColumn('Name', renderer, text=SND_NAME)
        self.sounds_treeview.append_column(column)

        renderer = Gtk.CellRendererText()
        renderer.sound_path = SND_PATH
        column = Gtk.TreeViewColumn('Path', renderer, text=SND_PATH)
        column.set_property('visible', False)
        self.sounds_treeview.append_column(column)

        self.sounds_treeview.set_model(self.sounds_model)

    def build_notifications_model_populate_treeview(self):
        # Notification Subscriptions treeview/model
        self.subscriptions_treeview = self.builder.get_object('subscriptions_treeview')
        subscriptions_selection = self.subscriptions_treeview.get_selection()
        subscriptions_selection.connect(
            'changed', self.on_subscriptions_treeview_selection_changed
        )
        self.subscriptions_treeview.set_tooltip_column(SUB_EVENT_DOC)
        self.subscriptions_model = Gtk.ListStore(str, str, bool, bool, bool, bool)

        renderer = Gtk.CellRendererText()
        setattr(renderer, 'event', SUB_EVENT)
        column = Gtk.TreeViewColumn('Event', renderer, text=SUB_EVENT)
        column.set_expand(True)
        self.subscriptions_treeview.append_column(column)

        renderer = Gtk.CellRendererText()
        setattr(renderer, 'event_doc', SUB_EVENT)
        column = Gtk.TreeViewColumn('Doc', renderer, text=SUB_EVENT_DOC)
        column.set_property('visible', False)
        self.subscriptions_treeview.append_column(column)

        renderer = Gtk.CellRendererToggle()
        renderer.set_property('activatable', True)
        renderer.connect('toggled', self._on_email_col_toggled)
        column = Gtk.TreeViewColumn('Email', renderer, active=SUB_NOT_EMAIL)
        column.set_clickable(True)
        self.subscriptions_treeview.append_column(column)

        renderer = Gtk.CellRendererToggle()
        renderer.set_property('activatable', True)
        renderer.connect('toggled', self._on_popup_col_toggled)
        column = Gtk.TreeViewColumn('Popup', renderer, active=SUB_NOT_POPUP)
        column.set_clickable(True)
        self.subscriptions_treeview.append_column(column)

        renderer = Gtk.CellRendererToggle()
        renderer.set_property('activatable', True)
        renderer.connect('toggled', self._on_blink_col_toggled)
        column = Gtk.TreeViewColumn('Blink', renderer, active=SUB_NOT_BLINK)
        column.set_clickable(True)
        self.subscriptions_treeview.append_column(column)

        renderer = Gtk.CellRendererToggle()
        renderer.set_property('activatable', True)
        renderer.connect('toggled', self._on_sound_col_toggled)
        column = Gtk.TreeViewColumn('Sound', renderer, active=SUB_NOT_SOUND)
        column.set_clickable(True)
        self.subscriptions_treeview.append_column(column)
        self.subscriptions_treeview.set_model(self.subscriptions_model)

    def popuplate_what_needs_handled_events(
        self, handled_events, email_subscriptions=None
    ):
        if email_subscriptions is None:
            email_subscriptions = []
        self.populate_subscriptions(handled_events, email_subscriptions)
        self.populate_sounds(handled_events)

    def populate_sounds(self, handled_events):
        self.sounds_model.clear()
        for event_name, event_doc in handled_events:
            if event_name in self.config['custom_sounds']:
                snd_path = self.config['custom_sounds'][event_name]
            else:
                snd_path = self.config['sound_path']

            if snd_path:
                self.sounds_model.set(
                    self.sounds_model.append(),
                    SND_EVENT,
                    event_name,
                    SND_EVENT_DOC,
                    event_doc,
                    SND_NAME,
                    basename(snd_path),
                    SND_PATH,
                    snd_path,
                )

    def populate_subscriptions(self, handled_events, email_subscriptions=None):
        if email_subscriptions is None:
            email_subscriptions = []
        subscriptions_dict = self.config['subscriptions']
        self.subscriptions_model.clear()
        #        self.handled_events = handled_events
        for event_name, event_doc in handled_events:
            self.subscriptions_model.set(
                self.subscriptions_model.append(),
                SUB_EVENT,
                event_name,
                SUB_EVENT_DOC,
                event_doc,
                SUB_NOT_EMAIL,
                event_name in email_subscriptions,
                SUB_NOT_POPUP,
                event_name in subscriptions_dict['popup'],
                SUB_NOT_BLINK,
                event_name in subscriptions_dict['blink'],
                SUB_NOT_SOUND,
                event_name in subscriptions_dict['sound'],
            )

    def on_apply_prefs(self):
        log.debug('applying prefs for Notifications')

        current_popup_subscriptions = []
        current_blink_subscriptions = []
        current_sound_subscriptions = []
        current_email_subscriptions = []
        for event, doc, email, popup, blink, sound in self.subscriptions_model:
            if email:
                current_email_subscriptions.append(event)
            if popup:
                current_popup_subscriptions.append(event)
            if blink:
                current_blink_subscriptions.append(event)
            if sound:
                current_sound_subscriptions.append(event)

        old_sound_file = self.config['sound_path']
        new_sound_file = self.builder.get_object('sound_path').get_filename()
        log.debug(
            'Old Default sound file: %s New one: %s', old_sound_file, new_sound_file
        )
        custom_sounds = {}
        for event_name, event_doc, filename, filepath in self.sounds_model:
            log.debug('Custom sound for event "%s": %s', event_name, filename)
            if filepath == old_sound_file:
                continue
            custom_sounds[event_name] = filepath

        self.config.config.update(
            {
                'popup_enabled': self.builder.get_object('popup_enabled').get_active(),
                'blink_enabled': self.builder.get_object('blink_enabled').get_active(),
                'sound_enabled': self.builder.get_object('sound_enabled').get_active(),
                'sound_path': new_sound_file,
                'subscriptions': {
                    'popup': current_popup_subscriptions,
                    'blink': current_blink_subscriptions,
                    'sound': current_sound_subscriptions,
                },
                'custom_sounds': custom_sounds,
            }
        )
        self.config.save()

        core_config = {
            'smtp_enabled': self.builder.get_object('smtp_enabled').get_active(),
            'smtp_host': self.builder.get_object('smtp_host').get_text(),
            'smtp_port': self.builder.get_object('smtp_port').get_value(),
            'smtp_user': self.builder.get_object('smtp_user').get_text(),
            'smtp_pass': self.builder.get_object('smtp_pass').get_text(),
            'smtp_from': self.builder.get_object('smtp_from').get_text(),
            'smtp_tls': self.builder.get_object('smtp_tls').get_active(),
            'smtp_recipients': [
                dest[0] for dest in self.recipients_model if dest[0] != 'USER@HOST'
            ],
            'subscriptions': {'email': current_email_subscriptions},
        }

        client.notifications.set_config(core_config)
        client.notifications.get_config().addCallback(self.cb_get_config)

    def on_show_prefs(self):
        client.notifications.get_config().addCallback(self.cb_get_config)

    def cb_get_config(self, core_config):
        """Callback for on show_prefs."""
        self.builder.get_object('smtp_host').set_text(core_config['smtp_host'])
        self.builder.get_object('smtp_port').set_value(core_config['smtp_port'])
        self.builder.get_object('smtp_user').set_text(core_config['smtp_user'])
        self.builder.get_object('smtp_pass').set_text(core_config['smtp_pass'])
        self.builder.get_object('smtp_from').set_text(core_config['smtp_from'])
        self.builder.get_object('smtp_tls').set_active(core_config['smtp_tls'])
        self.recipients_model.clear()
        for recipient in core_config['smtp_recipients']:
            self.recipients_model.set(
                self.recipients_model.append(),
                RECIPIENT_FIELD,
                recipient,
                RECIPIENT_EDIT,
                False,
            )
        self.builder.get_object('smtp_enabled').set_active(core_config['smtp_enabled'])
        self.builder.get_object('sound_enabled').set_active(
            self.config['sound_enabled']
        )
        self.builder.get_object('popup_enabled').set_active(
            self.config['popup_enabled']
        )
        self.builder.get_object('blink_enabled').set_active(
            self.config['blink_enabled']
        )
        if self.config['sound_path']:
            sound_path = self.config['sound_path']
        else:
            sound_path = deluge.common.get_default_download_dir()
        self.builder.get_object('sound_path').set_filename(sound_path)
        # Force toggle
        self.on_enabled_toggled(self.builder.get_object('smtp_enabled'))
        self.on_sound_enabled_toggled(self.builder.get_object('sound_enabled'))

        client.notifications.get_handled_events().addCallback(
            self.popuplate_what_needs_handled_events,
            core_config['subscriptions']['email'],
        )

    def on_sound_path_update_preview(self, filechooser):
        client.notifications.get_handled_events().addCallback(self.populate_sounds)

    def on_add_button_clicked(self, widget, treeview):
        model = treeview.get_model()
        model.set(model.append(), RECIPIENT_FIELD, 'USER@HOST', RECIPIENT_EDIT, True)

    def on_delete_button_clicked(self, widget, treeview):
        selection = treeview.get_selection()
        model, selected_iter = selection.get_selected()
        if selected_iter:
            model.remove(selected_iter)

    def on_cell_edited(self, cell, path_string, new_text, model):
        str_iter = model.get_iter_from_string(path_string)
        model.set(str_iter, RECIPIENT_FIELD, new_text)

    def on_recipients_treeview_selection_changed(self, selection):
        model, selected_connection_iter = selection.get_selected()
        if selected_connection_iter:
            self.builder.get_object('delete_button').set_property('sensitive', True)
        else:
            self.builder.get_object('delete_button').set_property('sensitive', False)

    def on_subscriptions_treeview_selection_changed(self, selection):
        model, selected_connection_iter = selection.get_selected()
        if selected_connection_iter:
            self.builder.get_object('delete_button').set_property('sensitive', True)
        else:
            self.builder.get_object('delete_button').set_property('sensitive', False)

    def on_sounds_treeview_selection_changed(self, selection):
        model, selected_iter = selection.get_selected()
        if selected_iter:
            self.builder.get_object('sounds_edit_button').set_property(
                'sensitive', True
            )
            path = model.get(selected_iter, SND_PATH)[0]
            log.debug('Sound selection changed: %s', path)
            if path != self.config['sound_path']:
                self.builder.get_object('sounds_revert_button').set_property(
                    'sensitive', True
                )
            else:
                self.builder.get_object('sounds_revert_button').set_property(
                    'sensitive', False
                )
        else:
            self.builder.get_object('sounds_edit_button').set_property(
                'sensitive', False
            )
            self.builder.get_object('sounds_revert_button').set_property(
                'sensitive', False
            )

    def on_sounds_revert_button_clicked(self, widget):
        log.debug('on_sounds_revert_button_clicked')
        selection = self.sounds_treeview.get_selection()
        model, selected_iter = selection.get_selected()
        if selected_iter:
            log.debug('on_sounds_revert_button_clicked: got iter')
            model.set(
                selected_iter,
                SND_PATH,
                self.config['sound_path'],
                SND_NAME,
                basename(self.config['sound_path']),
            )

    def on_sounds_edit_button_clicked(self, widget):
        log.debug('on_sounds_edit_button_clicked')
        selection = self.sounds_treeview.get_selection()
        model, selected_iter = selection.get_selected()
        if selected_iter:
            path = model.get(selected_iter, SND_PATH)[0]
            dialog = Gtk.FileChooserDialog(
                title=_('Choose Sound File'),
                buttons=(
                    Gtk.STOCK_CANCEL,
                    Gtk.ResponseType.CANCEL,
                    Gtk.STOCK_OPEN,
                    Gtk.ResponseType.OK,
                ),
            )
            dialog.set_filename(path)

            def update_model(response):
                if response == Gtk.ResponseType.OK:
                    new_filename = dialog.get_filename()
                    dialog.destroy()
                    log.debug(new_filename)
                    model.set(
                        selected_iter,
                        SND_PATH,
                        new_filename,
                        SND_NAME,
                        basename(new_filename),
                    )

            d = defer.maybeDeferred(dialog.run)
            d.addCallback(update_model)

            log.debug('dialog should have been shown')

    def on_enabled_toggled(self, widget):
        for widget_name in (
            'smtp_host',
            'smtp_port',
            'smtp_user',
            'smtp_pass',
            'smtp_pass',
            'smtp_tls',
            'smtp_from',
            'smtp_recipients',
        ):
            self.builder.get_object(widget_name).set_property(
                'sensitive', widget.get_active()
            )

    def on_sound_enabled_toggled(self, widget):
        if widget.get_active():
            self.builder.get_object('sound_path').set_property('sensitive', True)
            self.builder.get_object('sounds_page').set_property('sensitive', True)
            self.builder.get_object('sounds_page_label').set_property('sensitive', True)
        else:
            self.builder.get_object('sound_path').set_property('sensitive', False)
            self.builder.get_object('sounds_page').set_property('sensitive', False)
            self.builder.get_object('sounds_page_label').set_property(
                'sensitive', False
            )

    #        for widget_name in ('sounds_path', 'sounds_page', 'sounds_page_label'):
    #            self.builder.get_object(widget_name).set_property('sensitive',
    #                                                            widget.get_active())

    def _on_email_col_toggled(self, cell, path):
        self.subscriptions_model[path][SUB_NOT_EMAIL] = not self.subscriptions_model[
            path
        ][SUB_NOT_EMAIL]
        return

    def _on_popup_col_toggled(self, cell, path):
        self.subscriptions_model[path][SUB_NOT_POPUP] = not self.subscriptions_model[
            path
        ][SUB_NOT_POPUP]
        return

    def _on_blink_col_toggled(self, cell, path):
        self.subscriptions_model[path][SUB_NOT_BLINK] = not self.subscriptions_model[
            path
        ][SUB_NOT_BLINK]
        return

    def _on_sound_col_toggled(self, cell, path):
        self.subscriptions_model[path][SUB_NOT_SOUND] = not self.subscriptions_model[
            path
        ][SUB_NOT_SOUND]
        return



#
# Copyright (C) 2009 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import hashlib
import logging
import os
import time

from twisted.internet.utils import getProcessOutputAndValue

import deluge.component as component
from deluge.common import windows_check
from deluge.configmanager import ConfigManager
from deluge.core.rpcserver import export
from deluge.event import DelugeEvent
from deluge.plugins.pluginbase import CorePluginBase

log = logging.getLogger(__name__)

DEFAULT_CONFIG = {'commands': []}

EXECUTE_ID = 0
EXECUTE_EVENT = 1
EXECUTE_COMMAND = 2

EVENT_MAP = {
    'complete': 'TorrentFinishedEvent',
    'added': 'TorrentAddedEvent',
    'removed': 'TorrentRemovedEvent',
}


class ExecuteCommandAddedEvent(DelugeEvent):
    """
    Emitted when a new command is added.
    """

    def __init__(self, command_id, event, command):
        self._args = [command_id, event, command]


class ExecuteCommandRemovedEvent(DelugeEvent):
    """
    Emitted when a command is removed.
    """

    def __init__(self, command_id):
        self._args = [command_id]


class Core(CorePluginBase):
    def enable(self):
        self.config = ConfigManager('execute.conf', DEFAULT_CONFIG)
        event_manager = component.get('EventManager')
        self.registered_events = {}
        self.preremoved_cache = {}

        # Go through the commands list and register event handlers
        for command in self.config['commands']:
            event = command[EXECUTE_EVENT]
            if event in self.registered_events:
                continue

            def create_event_handler(event):
                def event_handler(torrent_id, *arg):
                    self.execute_commands(torrent_id, event, *arg)

                return event_handler

            event_handler = create_event_handler(event)
            event_manager.register_event_handler(EVENT_MAP[event], event_handler)
            if event == 'removed':
                event_manager.register_event_handler(
                    'PreTorrentRemovedEvent', self.on_preremoved
                )
            self.registered_events[event] = event_handler

        log.debug('Execute core plugin enabled!')

    def on_preremoved(self, torrent_id):
        # Get and store the torrent info before it is removed
        torrent = component.get('TorrentManager').torrents[torrent_id]
        info = torrent.get_status(['name', 'download_location'])
        self.preremoved_cache[torrent_id] = [
            torrent_id,
            info['name'],
            info['download_location'],
        ]

    def execute_commands(self, torrent_id, event, *arg):
        if event == 'added' and arg[0]:
            # No futher action as from_state (arg[0]) is True
            return
        elif event == 'removed':
            torrent_id, torrent_name, download_location = self.preremoved_cache.pop(
                torrent_id
            )
        else:
            torrent = component.get('TorrentManager').torrents[torrent_id]
            info = torrent.get_status(['name', 'download_location'])
            # Grab the torrent name and download location
            # getProcessOutputAndValue requires args to be str
            torrent_name = info['name']
            download_location = info['download_location']

        log.debug('Running commands for %s', event)

        def log_error(result, command):
            (stdout, stderr, exit_code) = result
            if exit_code:
                log.warning('Command "%s" failed with exit code %d', command, exit_code)
                if stdout:
                    log.warning('stdout: %s', stdout)
                if stderr:
                    log.warning('stderr: %s', stderr)

        # Go through and execute all the commands
        for command in self.config['commands']:
            if command[EXECUTE_EVENT] == event:
                command = os.path.expandvars(command[EXECUTE_COMMAND])
                command = os.path.expanduser(command)

                cmd_args = [
                    torrent_id.encode('utf8'),
                    torrent_name.encode('utf8'),
                    download_location.encode('utf8'),
                ]
                if windows_check():
                    # Escape ampersand on windows (see #2784)
                    cmd_args = [cmd_arg.replace(b'&', b'^^^&') for cmd_arg in cmd_args]

                if os.path.isfile(command) and os.access(command, os.X_OK):
                    log.debug('Running %s with args: %s', command, cmd_args)
                    d = getProcessOutputAndValue(command, cmd_args, env=os.environ)
                    d.addCallback(log_error, command)
                else:
                    log.error('Execute script not found or not executable')

    def disable(self):
        self.config.save()
        event_manager = component.get('EventManager')
        for event, handler in self.registered_events.items():
            event_manager.deregister_event_handler(event, handler)
        log.debug('Execute core plugin disabled!')

    # Exported RPC methods #
    @export
    def add_command(self, event, command):
        command_id = hashlib.sha1(str(time.time()).encode()).hexdigest()
        self.config['commands'].append((command_id, event, command))
        self.config.save()
        component.get('EventManager').emit(
            ExecuteCommandAddedEvent(command_id, event, command)
        )

    @export
    def get_commands(self):
        return self.config['commands']

    @export
    def remove_command(self, command_id):
        for command in self.config['commands']:
            if command[EXECUTE_ID] == command_id:
                self.config['commands'].remove(command)
                component.get('EventManager').emit(
                    ExecuteCommandRemovedEvent(command_id)
                )
                break
        self.config.save()

    @export
    def save_command(self, command_id, event, cmd):
        for i, command in enumerate(self.config['commands']):
            if command[EXECUTE_ID] == command_id:
                self.config['commands'][i] = (command_id, event, cmd)
                break
        self.config.save()



#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import argparse

import pytest

from deluge.ui.console.cmdline.commands.add import Command
from deluge.ui.console.cmdline.commands.config import json_eval
from deluge.ui.console.widgets.fields import TextInput


class MockParent:
    def __init__(self):
        self.border_off_x = 1
        self.pane_width = 20
        self.encoding = 'utf8'


class TestUIConsoleField:
    @pytest.fixture(autouse=True)
    def set_up(self):
        self.parent = MockParent()

    def test_text_input(self):
        def move_func(self, r, c):
            self._cursor_row = r
            self._cursor_col = c

        t = TextInput(
            self.parent,
            'name',
            'message',
            move_func,
            20,
            '/text/field/file/path',
            complete=False,
        )
        assert t
        assert t.handle_read(33)


class TestUIConsoleCommands:
    def test_add_move_completed(self):
        completed_path = 'completed_path'
        parser = argparse.ArgumentParser()
        cmd = Command()
        cmd.add_arguments(parser)
        args = parser.parse_args(['torrent', '-m', completed_path])
        assert args.move_completed_path == completed_path
        args = parser.parse_args(['torrent', '--move-path', completed_path])
        assert args.move_completed_path == completed_path

    def test_config_json_eval(self):
        assert json_eval('/downloads') == '/downloads'
        assert json_eval('/dir/with space') == '/dir/with space'
        assert json_eval('c:\\\\downloads') == 'c:\\\\downloads'
        assert json_eval('c:/downloads') == 'c:/downloads'
        # Ensure newlines are split and only first setting is used.
        assert json_eval('setting\nwithneline') == 'setting'
        # Allow both parentheses and square brackets.
        assert json_eval('(8000, 8001)') == [8000, 8001]
        assert json_eval('[8000, 8001]') == [8000, 8001]
        assert json_eval('["abc", "def"]') == ['abc', 'def']
        assert json_eval('{"foo": "bar"}') == {'foo': 'bar'}
        assert json_eval('{"number": 1234}') == {'number': 1234}
        # Hex string for peer_tos.
        assert json_eval('0x00') == '0x00'
        assert json_eval('1000') == 1000
        assert json_eval('-6') == -6
        assert json_eval('10.5') == 10.5
        assert json_eval('True')
        assert not json_eval('false')
        assert json_eval('none') is None
        # Empty values to clear config key.
        assert json_eval('[]') == []
        assert json_eval('') == ''



#
# Copyright (C) 2010 John Garland <johnnybg+deluge@gmail.com>
#
# Basic plugin template created by:
# Copyright (C) 2008 Martijn Voncken <mvoncken@gmail.com>
# Copyright (C) 2007-2009 Andrew Resch <andrewresch@gmail.com>
# Copyright (C) 2009 Damien Churchill <damoxc@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging

import deluge.component as component
from deluge.plugins.pluginbase import Gtk3PluginBase
from deluge.ui.client import client

log = logging.getLogger(__name__)


class GtkUI(Gtk3PluginBase):
    def enable(self):
        self.core = client.toggle
        self.plugin = component.get('PluginManager')
        self.separator = self.plugin.add_toolbar_separator()
        self.button = self.plugin.add_toolbar_button(
            self._on_button_clicked,
            label='Pause Session',
            stock='gtk-media-pause',
            tooltip='Pause the session',
        )

    def disable(self):
        component.get('PluginManager').remove_toolbar_button(self.button)
        component.get('PluginManager').remove_toolbar_button(self.separator)

    def update(self):
        def _on_get_status(paused):
            if paused:
                self.button.set_label('Resume Session')
                self.button.set_tooltip_text('Resume the session')
                self.button.set_stock_id('gtk-media-play')
            else:
                self.button.set_label('Pause Session')
                self.button.set_tooltip_text('Pause the session')
                self.button.set_stock_id('gtk-media-pause')

        self.core.get_status().addCallback(_on_get_status)

    def _on_button_clicked(self, widget):
        self.core.toggle()



#
# Copyright (C) 2008 Martijn Voncken <mvoncken@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#


import logging

from gi.repository.Gtk import Menu, MenuItem

from deluge import component  # for systray
from deluge.ui.client import client

log = logging.getLogger(__name__)


# Deferred Translation
def _(message):
    return message


NO_LABEL = _('No Label')
del _


class LabelMenu(MenuItem):
    def __init__(self):
        MenuItem.__init__(self, _('Label'))  # noqa: F821

        self.sub_menu = Menu()
        self.set_submenu(self.sub_menu)
        self.items = []

        # attach..
        self.sub_menu.connect('show', self.on_show, None)

    def get_torrent_ids(self):
        return component.get('TorrentView').get_selected_torrents()

    def on_show(self, widget=None, data=None):
        log.debug('label-on-show')
        client.label.get_labels().addCallback(self.cb_labels)

    def cb_labels(self, labels):
        for child in self.sub_menu.get_children():
            self.sub_menu.remove(child)
        for label in [NO_LABEL] + list(labels):
            if label == NO_LABEL:
                item = MenuItem(_(NO_LABEL))  # noqa: F821
            else:
                item = MenuItem(label.replace('_', '__'))
            item.connect('activate', self.on_select_label, label)
            self.sub_menu.append(item)
        self.show_all()

    def on_select_label(self, widget=None, label_id=None):
        log.debug('select label:%s,%s', label_id, self.get_torrent_ids())
        for torrent_id in self.get_torrent_ids():
            client.label.set_torrent(torrent_id, label_id)



#
# Copyright (C) 2009 Andrew Resch <andrewresch@gmail.com>
# Copyright (C) 2011 Pedro Algarvio <pedro@algarvio.me>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging
import subprocess
import sys

from twisted.internet import defer, reactor, ssl
from twisted.internet.protocol import ClientFactory

from deluge import error
from deluge.common import get_localhost_auth, get_version
from deluge.decorators import deprecated
from deluge.transfer import DelugeTransferProtocol

RPC_RESPONSE = 1
RPC_ERROR = 2
RPC_EVENT = 3

log = logging.getLogger(__name__)


def format_kwargs(kwargs):
    return ', '.join([key + '=' + str(value) for key, value in kwargs.items()])


class DelugeRPCRequest:
    """
    This object is created whenever there is a RPCRequest to be sent to the
    daemon.  It is generally only used by the DaemonProxy's call method.
    """

    request_id = None
    method = None
    args = None
    kwargs = None

    def __repr__(self):
        """
        Returns a string of the RPCRequest in the following form:
            method(arg, kwarg=foo, ...)
        """
        s = self.method + '('
        if self.args:
            s += ', '.join([str(x) for x in self.args])
        if self.kwargs:
            if self.args:
                s += ', '
            s += format_kwargs(self.kwargs)
        s += ')'

        return s

    def format_message(self):
        """
        Returns a properly formatted RPCRequest based on the properties.  Will
        raise a TypeError if the properties haven't been set yet.

        :returns: a properly formatted RPCRequest
        """
        if (
            self.request_id is None
            or self.method is None
            or self.args is None
            or self.kwargs is None
        ):
            raise TypeError(
                'You must set the properties of this object before calling format_message!'
            )

        return (self.request_id, self.method, self.args, self.kwargs)


class DelugeRPCProtocol(DelugeTransferProtocol):
    def connectionMade(self):  # NOQA: N802
        self.__rpc_requests = {}
        # Set the protocol in the daemon so it can send data
        self.factory.daemon.protocol = self
        # Get the address of the daemon that we've connected to
        peer = self.transport.getPeer()
        self.factory.daemon.host = peer.host
        self.factory.daemon.port = peer.port
        self.factory.daemon.connected = True
        log.debug('Connected to daemon at %s:%s..', peer.host, peer.port)
        self.factory.daemon.connect_deferred.callback((peer.host, peer.port))

    def message_received(self, request):
        """
        This method is called whenever we receive a message from the daemon.

        :param request: a tuple that should be either a RPCResponse, RCPError or RPCSignal

        """
        if not isinstance(request, tuple):
            log.debug('Received invalid message: type is not tuple')
            return
        if len(request) < 3:
            log.debug(
                'Received invalid message: number of items in ' 'response is %s',
                len(request),
            )
            return

        message_type = request[0]

        if message_type == RPC_EVENT:
            event = request[1]
            # log.debug('Received RPCEvent: %s', event)
            # A RPCEvent was received from the daemon so run any handlers
            # associated with it.
            if event in self.factory.event_handlers:
                for handler in self.factory.event_handlers[event]:
                    reactor.callLater(0, handler, *request[2])
            return

        request_id = request[1]

        # We get the Deferred object for this request_id to either run the
        # callbacks or the errbacks dependent on the response from the daemon.
        d = self.factory.daemon.pop_deferred(request_id)

        if message_type == RPC_RESPONSE:
            # Run the callbacks registered with this Deferred object
            d.callback(request[2])
        elif message_type == RPC_ERROR:
            # Recreate exception and errback'it
            try:
                # The exception class is located in deluge.error
                try:
                    exception_cls = getattr(error, request[2])
                    exception = exception_cls(*request[3], **request[4])
                except TypeError:
                    log.warning(
                        'Received invalid RPC_ERROR (Old daemon?): %s', request[2]
                    )
                    return

                # Ideally we would chain the deferreds instead of instance
                # checking just to log them. But, that would mean that any
                # errback on the fist deferred should returns it's failure
                # so it could pass back to the 2nd deferred on the chain. But,
                # that does not always happen.
                # So, just do some instance checking and just log rpc error at
                # different levels.
                r = self.__rpc_requests[request_id]
                msg = 'RPCError Message Received!'
                msg += '\n' + '-' * 80
                msg += '\n' + 'RPCRequest: ' + r.__repr__()
                msg += '\n' + '-' * 80
                if isinstance(exception, error.WrappedException):
                    msg += '\n' + exception.type + '\n' + exception.message + ': '
                    msg += exception.traceback
                else:
                    msg += '\n' + request[5] + '\n' + request[2] + ': '
                    msg += str(exception)
                msg += '\n' + '-' * 80

                if not isinstance(exception, error._ClientSideRecreateError):
                    # Let's log these as errors
                    log.error(msg)
                else:
                    # The rest just gets logged in debug level, just to log
                    # what's happening
                    log.debug(msg)
            except Exception:
                import traceback

                log.error(
                    'Failed to handle RPC_ERROR (Old daemon?): %s\nLocal error: %s',
                    request[2],
                    traceback.format_exc(),
                )
            d.errback(exception)
        del self.__rpc_requests[request_id]

    def send_request(self, request):
        """
        Sends a RPCRequest to the server.

        :param request: RPCRequest

        """
        try:
            # Store the DelugeRPCRequest object just in case a RPCError is sent in
            # response to this request.  We use the extra information when printing
            # out the error for debugging purposes.
            self.__rpc_requests[request.request_id] = request
            # log.debug('Sending RPCRequest %s: %s', request.request_id, request)
            # Send the request in a tuple because multiple requests can be sent at once
            self.transfer_message((request.format_message(),))
        except Exception as ex:
            log.warning('Error occurred when sending message: %s', ex)


class DelugeRPCClientFactory(ClientFactory):
    protocol = DelugeRPCProtocol

    def __init__(self, daemon, event_handlers):
        self.daemon = daemon
        self.event_handlers = event_handlers

    def startedConnecting(self, connector):  # NOQA: N802
        log.debug('Connecting to daemon at "%s:%s"...', connector.host, connector.port)

    def clientConnectionFailed(self, connector, reason):  # NOQA: N802
        log.debug(
            'Connection to daemon at "%s:%s" failed: %s',
            connector.host,
            connector.port,
            reason.value,
        )
        self.daemon.connect_deferred.errback(reason)

    def clientConnectionLost(self, connector, reason):  # NOQA: N802
        log.debug(
            'Connection lost to daemon at "%s:%s" reason: %s',
            connector.host,
            connector.port,
            reason.value,
        )
        self.daemon.host = None
        self.daemon.port = None
        self.daemon.username = None
        self.daemon.connected = False

        if (
            self.daemon.disconnect_deferred
            and not self.daemon.disconnect_deferred.called
        ):
            self.daemon.disconnect_deferred.callback(reason.value)
            self.daemon.disconnect_deferred = None

        if self.daemon.disconnect_callback:
            self.daemon.disconnect_callback()


class DaemonProxy:
    pass


class DaemonSSLProxy(DaemonProxy):
    def __init__(self, event_handlers=None):
        if event_handlers is None:
            event_handlers = {}
        self.__factory = DelugeRPCClientFactory(self, event_handlers)
        self.__factory.noisy = False
        self.__request_counter = 0
        self.__deferred = {}

        # This is set when a connection is made to the daemon
        self.protocol = None

        # This is set when a connection is made
        self.host = None
        self.port = None
        self.username = None
        self.authentication_level = 0

        self.connected = False

        self.disconnect_deferred = None
        self.disconnect_callback = None

        self.auth_levels_mapping = None
        self.auth_levels_mapping_reverse = None

    def connect(self, host, port):
        """
        Connects to a daemon at host:port

        :param host: str, the host to connect to
        :param port: int, the listening port on the daemon

        :returns: twisted.Deferred

        """
        log.debug('sslproxy.connect()')
        self.host = host
        self.port = port
        self.__connector = reactor.connectSSL(
            self.host, self.port, self.__factory, ssl.ClientContextFactory()
        )
        self.connect_deferred = defer.Deferred()
        self.daemon_info_deferred = defer.Deferred()

        # Upon connect we do a 'daemon.login' RPC
        self.connect_deferred.addCallback(self.__on_connect)
        self.connect_deferred.addErrback(self.__on_connect_fail)

        return self.daemon_info_deferred

    def disconnect(self):
        log.debug('sslproxy.disconnect()')
        self.disconnect_deferred = defer.Deferred()
        self.__connector.disconnect()
        return self.disconnect_deferred

    def call(self, method, *args, **kwargs):
        """
        Makes a RPCRequest to the daemon.  All methods should be in the form of
        'component.method'.

        :params method: str, the method to call in the form of 'component.method'
        :params args: the arguments to call the remote method with
        :params kwargs: the keyword arguments to call the remote method with

        :return: a twisted.Deferred object that will be activated when a RPCResponse
            or RPCError is received from the daemon

        """
        # Create the DelugeRPCRequest to pass to protocol.send_request()
        request = DelugeRPCRequest()
        request.request_id = self.__request_counter
        request.method = method
        request.args = args
        request.kwargs = kwargs
        # Send the request to the server
        self.protocol.send_request(request)
        # Create a Deferred object to return and add a default errback to print
        # the error.
        d = defer.Deferred()

        # Store the Deferred until we receive a response from the daemon
        self.__deferred[self.__request_counter] = d

        # Increment the request counter since we don't want to use the same one
        # before a response is received.
        self.__request_counter += 1

        return d

    def pop_deferred(self, request_id):
        """
        Pops a Deferred object.  This is generally called once we receive the
        reply we were waiting on from the server.

        :param request_id: the request_id of the Deferred to pop
        :type request_id: int

        """
        return self.__deferred.pop(request_id)

    def register_event_handler(self, event, handler):
        """
        Registers a handler function to be called when `:param:event` is received
        from the daemon.

        :param event: the name of the event to handle
        :type event: str
        :param handler: the function to be called when `:param:event`
            is emitted from the daemon
        :type handler: function

        """
        if event not in self.__factory.event_handlers:
            # This is a new event to handle, so we need to tell the daemon
            # that we're interested in receiving this type of event
            self.__factory.event_handlers[event] = []
            if self.connected:
                self.call('daemon.set_event_interest', [event])

        # Only add the handler if it's not already registered
        if handler not in self.__factory.event_handlers[event]:
            self.__factory.event_handlers[event].append(handler)

    def deregister_event_handler(self, event, handler):
        """
        Deregisters a event handler.

        :param event: the name of the event
        :type event: str
        :param handler: the function registered
        :type handler: function

        """
        if (
            event in self.__factory.event_handlers
            and handler in self.__factory.event_handlers[event]
        ):
            self.__factory.event_handlers[event].remove(handler)

    def __on_connect(self, result):
        log.debug('__on_connect called')

        def on_info(daemon_info):
            self.daemon_info = daemon_info
            log.debug('Got info from daemon: %s', daemon_info)
            self.daemon_info_deferred.callback(daemon_info)

        def on_info_fail(reason):
            log.debug('Failed to get info from daemon')
            log.exception(reason)
            self.daemon_info_deferred.errback(reason)

        self.call('daemon.info').addCallback(on_info).addErrback(on_info_fail)
        return self.daemon_info_deferred

    def __on_connect_fail(self, reason):
        self.daemon_info_deferred.errback(reason)

    def authenticate(self, username, password):
        log.debug('%s.authenticate: %s', self.__class__.__name__, username)
        login_deferred = defer.Deferred()
        d = self.call('daemon.login', username, password, client_version=get_version())
        d.addCallbacks(
            self.__on_login,
            self.__on_login_fail,
            callbackArgs=[username, login_deferred],
            errbackArgs=[login_deferred],
        )
        return login_deferred

    def __on_login(self, result, username, login_deferred):
        log.debug('__on_login called: %s %s', username, result)
        self.username = username
        self.authentication_level = result
        # We need to tell the daemon what events we're interested in receiving
        if self.__factory.event_handlers:
            self.call('daemon.set_event_interest', list(self.__factory.event_handlers))
            self.call('core.get_auth_levels_mappings').addCallback(
                self.__on_auth_levels_mappings
            )

        login_deferred.callback(result)

    def __on_login_fail(self, result, login_deferred):
        login_deferred.errback(result)

    def __on_auth_levels_mappings(self, result):
        auth_levels_mapping, auth_levels_mapping_reverse = result
        self.auth_levels_mapping = auth_levels_mapping
        self.auth_levels_mapping_reverse = auth_levels_mapping_reverse

    def set_disconnect_callback(self, cb):
        """
        Set a function to be called when the connection to the daemon is lost
        for any reason.
        """
        self.disconnect_callback = cb

    def get_bytes_recv(self):
        return self.protocol.get_bytes_recv()

    def get_bytes_sent(self):
        return self.protocol.get_bytes_sent()


class DaemonStandaloneProxy(DaemonProxy):
    def __init__(self, event_handlers=None):
        if event_handlers is None:
            event_handlers = {}
        from deluge.core import daemon

        self.__daemon = daemon.Daemon(standalone=True)
        self.__daemon.start()
        log.debug('daemon created!')
        self.connected = True
        self.host = 'localhost'
        self.port = 58846
        # Running in standalone mode, it's safe to import auth level
        from deluge.core.authmanager import (
            AUTH_LEVEL_ADMIN,
            AUTH_LEVELS_MAPPING,
            AUTH_LEVELS_MAPPING_REVERSE,
        )

        self.username = 'localclient'
        self.authentication_level = AUTH_LEVEL_ADMIN
        self.auth_levels_mapping = AUTH_LEVELS_MAPPING
        self.auth_levels_mapping_reverse = AUTH_LEVELS_MAPPING_REVERSE
        # Register the event handlers
        for event in event_handlers:
            for handler in event_handlers[event]:
                self.__daemon.core.eventmanager.register_event_handler(event, handler)

    def disconnect(self):
        self.connected = False
        self.__daemon = None

    def call(self, method, *args, **kwargs):
        # log.debug('call: %s %s %s', method, args, kwargs)

        import copy

        try:
            m = self.__daemon.rpcserver.get_object_method(method)
        except Exception as ex:
            log.exception(ex)
            return defer.fail(ex)
        else:
            return defer.maybeDeferred(m, *copy.deepcopy(args), **copy.deepcopy(kwargs))

    def register_event_handler(self, event, handler):
        """
        Registers a handler function to be called when `:param:event` is
        received from the daemon.

        :param event: the name of the event to handle
        :type event: str
        :param handler: the function to be called when `:param:event`
            is emitted from the daemon
        :type handler: function

        """
        self.__daemon.core.eventmanager.register_event_handler(event, handler)

    def deregister_event_handler(self, event, handler):
        """
        Deregisters a event handler.

        :param event: the name of the event
        :type event: str
        :param handler: the function registered
        :type handler: function

        """
        self.__daemon.core.eventmanager.deregister_event_handler(event, handler)


class DottedObject:
    """
    This is used for dotted name calls to client
    """

    def __init__(self, daemon, method):
        self.daemon = daemon
        self.base = method

    def __call__(self, *args, **kwargs):
        raise Exception('You must make calls in the form of "component.method"')

    def __getattr__(self, name):
        return RemoteMethod(self.daemon, self.base + '.' + name)


class RemoteMethod(DottedObject):
    """
    This is used when something like 'client.core.get_something()' is attempted.
    """

    def __call__(self, *args, **kwargs):
        return self.daemon.call(self.base, *args, **kwargs)


class Client:
    """
    This class is used to connect to a daemon process and issue RPC requests.
    """

    __event_handlers = {}

    def __init__(self):
        self._daemon_proxy = None
        self.disconnect_callback = None
        self.__started_standalone = False

    def connect(
        self,
        host='127.0.0.1',
        port=58846,
        username='',
        password='',
        skip_authentication=False,
    ):
        """
        Connects to a daemon process.

        :param host: str, the hostname of the daemon
        :param port: int, the port of the daemon
        :param username: str, the username to login with
        :param password: str, the password to login with

        :returns: a Deferred object that will be called once the connection
            has been established or fails
        """

        self._daemon_proxy = DaemonSSLProxy(dict(self.__event_handlers))
        self._daemon_proxy.set_disconnect_callback(self.__on_disconnect)

        d = self._daemon_proxy.connect(host, port)

        def on_connected(daemon_version):
            log.debug('on_connected. Daemon version: %s', daemon_version)
            return daemon_version

        def on_connect_fail(reason):
            log.debug('on_connect_fail: %s', reason)
            self.disconnect()
            return reason

        def on_authenticate(result, daemon_info):
            log.debug('Authentication successful: %s', result)
            return result

        def on_authenticate_fail(reason):
            log.debug('Failed to authenticate: %s', reason.value)
            return reason

        def authenticate(daemon_version, username, password):
            if not username and host in ('127.0.0.1', 'localhost'):
                # No username provided and it's localhost, so attempt to get credentials from auth file.
                username, password = get_localhost_auth()

            d = self._daemon_proxy.authenticate(username, password)
            d.addCallback(on_authenticate, daemon_version)
            d.addErrback(on_authenticate_fail)
            return d

        d.addCallbacks(on_connected)
        d.addErrback(on_connect_fail)
        if not skip_authentication:
            d.addCallback(authenticate, username, password)
        return d

    def disconnect(self):
        """
        Disconnects from the daemon.
        """
        if self.is_standalone():
            self._daemon_proxy.disconnect()
            self.stop_standalone()
            return defer.succeed(True)

        if self._daemon_proxy:
            return self._daemon_proxy.disconnect()

    def start_standalone(self):
        """
        Starts a daemon in the same process as the client.
        """
        self._daemon_proxy = DaemonStandaloneProxy(self.__event_handlers)
        self.__started_standalone = True

    def stop_standalone(self):
        """
        Stops the daemon process in the client.
        """
        self._daemon_proxy = None
        self.__started_standalone = False

    @deprecated
    def start_classic_mode(self):
        """Deprecated: Use start_standalone"""
        self.start_standalone()

    @deprecated
    def stop_classic_mode(self):
        """Deprecated: Use stop_standalone"""
        self.stop_standalone()

    def start_daemon(self, port, config):
        """Starts a daemon process.

        Args:
            port (int): Port for the daemon to listen on.
            config (str): Config path to pass to daemon.

        Returns:
            bool: True is successfully started the daemon, False otherwise.

        """
        # subprocess.popen does not work with unicode args (with non-ascii characters) on windows
        config = config.encode(sys.getfilesystemencoding())
        try:
            subprocess.Popen(['deluged', '--port=%s' % port, b'--config=%s' % config])
        except OSError as ex:
            from errno import ENOENT

            if ex.errno == ENOENT:
                log.error(
                    _(
                        'Deluge cannot find the `deluged` executable, check that '
                        'the deluged package is installed, or added to your PATH.'
                    )
                )
            else:
                log.exception(ex)
        except Exception as ex:
            log.error('Unable to start daemon!')
            log.exception(ex)
        else:
            return True
        return False

    def is_localhost(self):
        """
        Checks if the current connected host is a localhost or not.

        :returns: bool, True if connected to a localhost

        """
        if (
            self._daemon_proxy
            and self._daemon_proxy.host in ('127.0.0.1', 'localhost')
            or isinstance(self._daemon_proxy, DaemonStandaloneProxy)
        ):
            return True
        return False

    def is_standalone(self):
        """
        Checks to see if the client has been started in standalone mode.

        :returns: bool, True if in standalone mode

        """
        return self.__started_standalone

    @deprecated
    def is_classicmode(self):
        """Deprecated: Use is_standalone"""
        self.is_standalone()

    def connected(self):
        """
        Check to see if connected to a daemon.

        :returns: bool, True if connected

        """
        return self._daemon_proxy.connected if self._daemon_proxy else False

    def connection_info(self):
        """
        Get some info about the connection or return None if not connected.

        :returns: a tuple of (host, port, username) or None if not connected
        """
        if self.connected():
            return (
                self._daemon_proxy.host,
                self._daemon_proxy.port,
                self._daemon_proxy.username,
            )

        return None

    def register_event_handler(self, event, handler):
        """
        Registers a handler that will be called when an event is received from the daemon.

        :params event: str, the event to handle
        :params handler: func, the handler function, f(args)
        """
        if event not in self.__event_handlers:
            self.__event_handlers[event] = []
        self.__event_handlers[event].append(handler)
        # We need to replicate this in the daemon proxy
        if self._daemon_proxy:
            self._daemon_proxy.register_event_handler(event, handler)

    def deregister_event_handler(self, event, handler):
        """
        Deregisters a event handler.

        :param event: str, the name of the event
        :param handler: function, the function registered

        """
        if event in self.__event_handlers and handler in self.__event_handlers[event]:
            self.__event_handlers[event].remove(handler)
        if self._daemon_proxy:
            self._daemon_proxy.deregister_event_handler(event, handler)

    def force_call(self, block=False):
        # no-op for now.. we'll see if we need this in the future
        pass

    def __getattr__(self, method):
        return DottedObject(self._daemon_proxy, method)

    def set_disconnect_callback(self, cb):
        """
        Set a function to be called whenever the client is disconnected from
        the daemon for any reason.
        """
        self.disconnect_callback = cb

    def __on_disconnect(self):
        if self.disconnect_callback:
            self.disconnect_callback()

    def get_bytes_recv(self):
        """
        Returns the number of bytes received from the daemon.

        :returns: the number of bytes received
        :rtype: int
        """
        return self._daemon_proxy.get_bytes_recv()

    def get_bytes_sent(self):
        """
        Returns the number of bytes sent to the daemon.

        :returns: the number of bytes sent
        :rtype: int
        """
        return self._daemon_proxy.get_bytes_sent()

    def get_auth_user(self):
        """
        Returns the current authenticated username.

        :returns: the authenticated username
        :rtype: str
        """
        return self._daemon_proxy.username

    def get_auth_level(self):
        """
        Returns the authentication level the daemon returned upon authentication.

        :returns: the authentication level
        :rtype: int
        """
        return self._daemon_proxy.authentication_level

    @property
    def auth_levels_mapping(self):
        return self._daemon_proxy.auth_levels_mapping

    @property
    def auth_levels_mapping_reverse(self):
        return self._daemon_proxy.auth_levels_mapping_reverse


# This is the object clients will use
client = Client()



#
# Copyright (C) 2009 Andrew Resch <andrewresch@gmail.com>
#
# Basic plugin template created by:
# Copyright (C) 2008 Martijn Voncken <mvoncken@gmail.com>
# Copyright (C) 2007-2009 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging

import gi  # isort:skip (Required before Gtk import).

gi.require_version('Gtk', '3.0')

# isort:imports-thirdparty
from gi.repository import Gtk

# isort:imports-firstparty
import deluge.component as component
from deluge.plugins.pluginbase import Gtk3PluginBase
from deluge.ui.client import client

# isort:imports-localfolder
from .common import get_resource

log = logging.getLogger(__name__)


class GtkUI(Gtk3PluginBase):
    def enable(self):
        self.builder = Gtk.Builder()
        self.builder.add_from_file(get_resource('extractor_prefs.ui'))

        component.get('Preferences').add_page(
            _('Extractor'), self.builder.get_object('extractor_prefs_box')
        )
        component.get('PluginManager').register_hook(
            'on_apply_prefs', self.on_apply_prefs
        )
        component.get('PluginManager').register_hook(
            'on_show_prefs', self.on_show_prefs
        )
        self.on_show_prefs()

    def disable(self):
        component.get('Preferences').remove_page(_('Extractor'))
        component.get('PluginManager').deregister_hook(
            'on_apply_prefs', self.on_apply_prefs
        )
        component.get('PluginManager').deregister_hook(
            'on_show_prefs', self.on_show_prefs
        )
        del self.builder

    def on_apply_prefs(self):
        log.debug('applying prefs for Extractor')
        if client.is_localhost():
            path = self.builder.get_object('folderchooser_path').get_filename()
        else:
            path = self.builder.get_object('entry_path').get_text()

        config = {
            'extract_path': path,
            'use_name_folder': self.builder.get_object('chk_use_name').get_active(),
        }

        client.extractor.set_config(config)

    def on_show_prefs(self):
        if client.is_localhost():
            self.builder.get_object('folderchooser_path').show()
            self.builder.get_object('entry_path').hide()
        else:
            self.builder.get_object('folderchooser_path').hide()
            self.builder.get_object('entry_path').show()

        def on_get_config(config):
            if client.is_localhost():
                self.builder.get_object('folderchooser_path').set_current_folder(
                    config['extract_path']
                )
            else:
                self.builder.get_object('entry_path').set_text(config['extract_path'])

            self.builder.get_object('chk_use_name').set_active(
                config['use_name_folder']
            )

        client.extractor.get_config().addCallback(on_get_config)



#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import os
import shutil
import warnings
from base64 import b64encode
from unittest import mock

import pytest
import pytest_twisted
from twisted.internet import reactor, task

from deluge import component
from deluge.bencode import bencode
from deluge.conftest import BaseTestCase
from deluge.core.core import Core
from deluge.core.rpcserver import RPCServer
from deluge.error import InvalidTorrentError

from . import common

warnings.filterwarnings('ignore', category=RuntimeWarning)
warnings.resetwarnings()


class TestTorrentmanager(BaseTestCase):
    def set_up(self):
        self.rpcserver = RPCServer(listen=False)
        self.core = Core()
        self.core.config.config['lsd'] = False
        self.clock = task.Clock()
        self.tm = self.core.torrentmanager
        self.tm.callLater = self.clock.callLater
        return component.start()

    def tear_down(self):
        def on_shutdown(result):
            del self.rpcserver
            del self.core

        return component.shutdown().addCallback(on_shutdown)

    @pytest_twisted.inlineCallbacks
    def test_remove_torrent(self):
        filename = common.get_test_data_file('test.torrent')
        with open(filename, 'rb') as _file:
            filedump = _file.read()
        torrent_id = yield self.core.add_torrent_file_async(
            filename, b64encode(filedump), {}
        )
        assert self.tm.remove(torrent_id, False)

    @pytest_twisted.inlineCallbacks
    def test_remove_magnet(self):
        """Test remove magnet before received metadata and delete_copies is True"""
        magnet = 'magnet:?xt=urn:btih:ab570cdd5a17ea1b61e970bb72047de141bce173'
        options = {}
        self.core.config.config['copy_torrent_file'] = True
        self.core.config.config['del_copy_torrent_file'] = True
        torrent_id = yield self.core.add_torrent_magnet(magnet, options)
        assert self.tm.remove(torrent_id, False)

    @pytest_twisted.ensureDeferred
    async def test_prefetch_metadata(self):
        from deluge._libtorrent import lt

        with open(common.get_test_data_file('test.torrent'), 'rb') as _file:
            t_info = lt.torrent_info(lt.bdecode(_file.read()))
        mock_alert = mock.MagicMock()
        mock_alert.handle.info_hash = mock.MagicMock(
            return_value='ab570cdd5a17ea1b61e970bb72047de141bce173'
        )
        mock_alert.handle.get_torrent_info = mock.MagicMock(return_value=t_info)

        magnet = 'magnet:?xt=urn:btih:ab570cdd5a17ea1b61e970bb72047de141bce173'
        d = self.tm.prefetch_metadata(magnet, 30)
        # Make sure to use calllater, because the above prefetch call won't
        # actually start running until we await it.
        reactor.callLater(0, self.tm.on_alert_metadata_received, mock_alert)

        expected = (
            'ab570cdd5a17ea1b61e970bb72047de141bce173',
            b64encode(
                bencode(
                    {
                        b'piece length': 32768,
                        b'sha1': (
                            b'2\xce\xb6\xa8"\xd7\xf0\xd4\xbf\xdc^K\xba\x1bh'
                            b'\x9d\xc5\xb7\xac\xdd'
                        ),
                        b'name': b'azcvsupdater_2.6.2.jar',
                        b'private': 0,
                        b'pieces': (
                            b"\xdb\x04B\x05\xc3'\xdab\xb8su97\xa9u"
                            b'\xca<w\\\x1ef\xd4\x9b\x16\xa9}\xc0\x9f:\xfd'
                            b'\x97qv\x83\xa2"\xef\x9d7\x0by!\rl\xe5v\xb7'
                            b'\x18{\xf7/"P\xe9\x8d\x01D\x9e8\xbd\x16\xe3'
                            b'\xfb-\x9d\xaa\xbcM\x11\xba\x92\xfc\x13F\xf0'
                            b'\x1c\x86x+\xc8\xd0S\xa9\x90`\xa1\xe4\x82\xe8'
                            b'\xfc\x08\xf7\xe3\xe5\xf6\x85\x1c%\xe7%\n\xed'
                            b'\xc0\x1f\xa1;\x9a\xea\xcf\x90\x0c/F>\xdf\xdagA'
                            b'\xc42|\xda\x82\xf5\xa6b\xa1\xb8#\x80wI\xd8f'
                            b'\xf8\xbd\xacW\xab\xc3s\xe0\xbbw\xf2K\xbe\xee'
                            b'\xa8rG\xe1W\xe8\xb7\xc2i\xf3\xd8\xaf\x9d\xdc'
                            b'\xd0#\xf4\xc1\x12u\xcd\x0bE?:\xe8\x9c\x1cu'
                            b'\xabb(oj\r^\xd5\xd5A\x83\x88\x9a\xa1J\x1c?'
                            b'\xa1\xd6\x8c\x83\x9e&'
                        ),
                        b'length': 307949,
                        b'name.utf-8': b'azcvsupdater_2.6.2.jar',
                        b'ed2k': b'>p\xefl\xfa]\x95K\x1b^\xc2\\;;e\xb7',
                    }
                )
            ),
        )
        assert expected == await d

    @pytest_twisted.ensureDeferred
    async def test_prefetch_metadata_timeout(self):
        magnet = 'magnet:?xt=urn:btih:ab570cdd5a17ea1b61e970bb72047de141bce173'
        d = self.tm.prefetch_metadata(magnet, 30)
        self.clock.advance(30)
        result = await d
        expected = ('ab570cdd5a17ea1b61e970bb72047de141bce173', b'')
        assert result == expected

    @pytest.mark.todo
    def test_remove_torrent_false(self):
        """Test when remove_torrent returns False"""
        common.todo_test(self)

    def test_remove_invalid_torrent(self):
        with pytest.raises(InvalidTorrentError):
            self.tm.remove('torrentidthatdoesntexist')

    def test_open_state(self):
        """Open a state with a UTF-8 encoded torrent filename."""
        shutil.copy(
            common.get_test_data_file('utf8_filename_torrents.state'),
            os.path.join(self.config_dir, 'state', 'torrents.state'),
        )

        state = self.tm.open_state()
        assert len(state.torrents) == 1



#
# Copyright (C) 2008 Andrew Resch <andrewresch@gmail.com>
# Copyright (C) 2009-2010 John Garland <johnnybg+deluge@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging
import os
import shutil
import time
from datetime import datetime, timedelta
from email.utils import formatdate
from urllib.parse import urljoin

from twisted.internet import defer, threads
from twisted.internet.task import LoopingCall
from twisted.web import error

import deluge.component as component
import deluge.configmanager
from deluge.common import is_url
from deluge.core.rpcserver import export
from deluge.httpdownloader import download_file
from deluge.plugins.pluginbase import CorePluginBase

from .common import IP, BadIP
from .detect import UnknownFormatError, create_reader, detect_compression, detect_format
from .readers import ReaderParseError

# TODO: check return values for deferred callbacks
# TODO: review class attributes for redundancy

log = logging.getLogger(__name__)

DEFAULT_PREFS = {
    'url': '',
    'load_on_start': False,
    'check_after_days': 4,
    'list_compression': '',
    'list_type': '',
    'last_update': 0.0,
    'list_size': 0,
    'timeout': 180,
    'try_times': 3,
    'whitelisted': [],
}

# Constants
ALLOW_RANGE = 0
BLOCK_RANGE = 1


class Core(CorePluginBase):
    def enable(self):
        log.debug('Blocklist: Plugin enabled...')

        self.is_url = True
        self.is_downloading = False
        self.is_importing = False
        self.has_imported = False
        self.up_to_date = False
        self.need_to_resume_session = False
        self.num_whited = 0
        self.num_blocked = 0
        self.file_progress = 0.0

        self.core = component.get('Core')
        self.config = deluge.configmanager.ConfigManager(
            'blocklist.conf', DEFAULT_PREFS
        )
        if 'whitelisted' not in self.config:
            self.config['whitelisted'] = []

        self.reader = create_reader(
            self.config['list_type'], self.config['list_compression']
        )

        if not isinstance(self.config['last_update'], float):
            self.config.config['last_update'] = 0.0

        update_now = False
        if self.config['load_on_start']:
            self.pause_session()
            if self.config['last_update']:
                last_update = datetime.fromtimestamp(self.config['last_update'])
                check_period = timedelta(days=self.config['check_after_days'])
            if (
                not self.config['last_update']
                or last_update + check_period < datetime.now()
            ):
                update_now = True
            else:
                d = self.import_list(
                    deluge.configmanager.get_config_dir('blocklist.cache')
                )
                d.addCallbacks(self.on_import_complete, self.on_import_error)
                if self.need_to_resume_session:
                    d.addBoth(self.resume_session)

        # This function is called every 'check_after_days' days, to download
        # and import a new list if needed.
        self.update_timer = LoopingCall(self.check_import)
        if self.config['check_after_days'] > 0:
            self.update_timer.start(
                self.config['check_after_days'] * 24 * 60 * 60, update_now
            )

    def disable(self):
        self.config.save()
        log.debug('Reset IP filter')
        self.core.session.get_ip_filter().add_rule(
            '0.0.0.0', '255.255.255.255', ALLOW_RANGE
        )
        log.debug('Blocklist: Plugin disabled')

    def update(self):
        pass

    # Exported RPC methods #
    @export
    def check_import(self, force=False):
        """Imports latest blocklist specified by blocklist url.

        Args:
            force (bool, optional): Force the download/import, default is False.

        Returns:
            Deferred: A Deferred which fires when the blocklist has been imported.

        """
        if not self.config['url']:
            return

        # Reset variables
        self.filename = None
        self.force_download = force
        self.failed_attempts = 0
        self.auto_detected = False
        self.up_to_date = False
        if force:
            self.reader = None
        self.is_url = is_url(self.config['url'])

        # Start callback chain
        if self.is_url:
            d = self.download_list()
            d.addCallbacks(self.on_download_complete, self.on_download_error)
            d.addCallback(self.import_list)
        else:
            d = self.import_list(self.config['url'])
        d.addCallbacks(self.on_import_complete, self.on_import_error)
        if self.need_to_resume_session:
            d.addBoth(self.resume_session)

        return d

    @export
    def get_config(self):
        """Gets the blocklist config dictionary.

        Returns:
            dict: The config dictionary.

        """
        return self.config.config

    @export
    def set_config(self, config):
        """Sets the blocklist config.

        Args:
            config (dict): config to set.

        """
        needs_blocklist_import = False
        for key in config:
            if key == 'whitelisted':
                saved = set(self.config[key])
                update = set(config[key])
                diff = saved.symmetric_difference(update)
                if diff:
                    log.debug('Whitelist changed. Updating...')
                    added = update.intersection(diff)
                    removed = saved.intersection(diff)
                    if added:
                        for ip in added:
                            try:
                                ip = IP.parse(ip)
                                self.blocklist.add_rule(
                                    ip.address, ip.address, ALLOW_RANGE
                                )
                                saved.add(ip.address)
                                log.debug('Added %s to whitelisted', ip)
                                self.num_whited += 1
                            except BadIP as ex:
                                log.error('Bad IP: %s', ex)
                                continue
                    if removed:
                        needs_blocklist_import = True
                        for ip in removed:
                            try:
                                ip = IP.parse(ip)
                                saved.remove(ip.address)
                                log.debug('Removed %s from whitelisted', ip)
                            except BadIP as ex:
                                log.error('Bad IP: %s', ex)
                                continue

                self.config[key] = list(saved)
                continue
            elif key == 'check_after_days':
                if self.config[key] != config[key]:
                    self.config[key] = config[key]
                    update_now = False
                    if self.config['last_update']:
                        last_update = datetime.fromtimestamp(self.config['last_update'])
                        check_period = timedelta(days=self.config['check_after_days'])
                    if (
                        not self.config['last_update']
                        or last_update + check_period < datetime.now()
                    ):
                        update_now = True
                    if self.update_timer.running:
                        self.update_timer.stop()
                    if self.config['check_after_days'] > 0:
                        self.update_timer.start(
                            self.config['check_after_days'] * 24 * 60 * 60, update_now
                        )
                continue
            self.config[key] = config[key]

        if needs_blocklist_import:
            log.debug(
                'IP addresses were removed from the whitelist. Since we '
                'do not know if they were blocked before. Re-import '
                'current blocklist and re-add whitelisted.'
            )
            self.has_imported = False
            d = self.import_list(deluge.configmanager.get_config_dir('blocklist.cache'))
            d.addCallbacks(self.on_import_complete, self.on_import_error)

    @export
    def get_status(self):
        """Get the status of the plugin.

        Returns:
            dict: The status dict of the plugin.

        """
        status = {}
        if self.is_downloading:
            status['state'] = 'Downloading'
        elif self.is_importing:
            status['state'] = 'Importing'
        else:
            status['state'] = 'Idle'

        status['up_to_date'] = self.up_to_date
        status['num_whited'] = self.num_whited
        status['num_blocked'] = self.num_blocked
        status['file_progress'] = self.file_progress
        status['file_url'] = self.config['url']
        status['file_size'] = self.config['list_size']
        status['file_date'] = self.config['last_update']
        status['file_type'] = self.config['list_type']
        status['whitelisted'] = self.config['whitelisted']
        if self.config['list_compression']:
            status['file_type'] += ' (%s)' % self.config['list_compression']
        return status

    ####

    def update_info(self, blocklist):
        """Updates blocklist info.

        Args:
            blocklist (str): Path of blocklist.

        Returns:
            str: Path of blocklist.

        """
        log.debug('Updating blocklist info: %s', blocklist)
        self.config['last_update'] = time.time()
        self.config['list_size'] = os.path.getsize(blocklist)
        self.filename = blocklist
        return blocklist

    def download_list(self, url=None):
        """Downloads the blocklist specified by 'url' in the config.

        Args:
            url (str, optional): url to download from, defaults to config value.

        Returns:
            Deferred: a Deferred which fires once the blocklist has been downloaded.

        """

        def on_retrieve_data(data, current_length, total_length):
            if total_length:
                fp = current_length / total_length
                if fp > 1.0:
                    fp = 1.0
            else:
                fp = 0.0

            self.file_progress = fp

        import socket

        socket.setdefaulttimeout(self.config['timeout'])

        if not url:
            url = self.config['url']

        headers = {}
        if self.config['last_update'] and not self.force_download:
            headers['If-Modified-Since'] = formatdate(
                self.config['last_update'], usegmt=True
            )

        log.debug('Attempting to download blocklist %s', url)
        log.debug('Sending headers: %s', headers)
        self.is_downloading = True
        return download_file(
            url,
            deluge.configmanager.get_config_dir('blocklist.download'),
            on_retrieve_data,
            headers,
        )

    def on_download_complete(self, blocklist):
        """Runs any download clean up functions.

        Args:
            blocklist (str): Path of blocklist.

        Returns:
            Deferred: a Deferred which fires when clean up is done.

        """
        log.debug('Blocklist download complete: %s', blocklist)
        self.is_downloading = False
        return threads.deferToThread(self.update_info, blocklist)

    def on_download_error(self, f):
        """Recovers from download error.

        Args:
            f (Failure): Failure that occurred.

        Returns:
            Deferred or Failure: A Deferred if recovery was possible else original Failure.

        """
        self.is_downloading = False
        error_msg = f.getErrorMessage()
        d = f
        if f.check(error.PageRedirect):
            # Handle redirect errors
            location = urljoin(self.config['url'], error_msg.split(' to ')[1])
            if 'Moved Permanently' in error_msg:
                log.debug('Setting blocklist url to %s', location)
                self.config['url'] = location
            d = self.download_list(location)
            d.addCallbacks(self.on_download_complete, self.on_download_error)
        else:
            if 'Not Modified' in error_msg:
                log.debug('Blocklist is up-to-date!')
                self.up_to_date = True
                blocklist = deluge.configmanager.get_config_dir('blocklist.cache')
                d = threads.deferToThread(self.update_info, blocklist)
            else:
                log.warning('Blocklist download failed: %s', error_msg)
                if self.failed_attempts < self.config['try_times']:
                    log.debug(
                        'Try downloading blocklist again... (%s/%s)',
                        self.failed_attempts,
                        self.config['try_times'],
                    )
                    self.failed_attempts += 1
                    d = self.download_list()
                    d.addCallbacks(self.on_download_complete, self.on_download_error)
        return d

    def import_list(self, blocklist):
        """Imports the downloaded blocklist into the session.

        Args:
            blocklist (str): path of blocklist.

        Returns:
            Deferred: A Deferred that fires when the blocklist has been imported.

        """
        log.trace('on import_list')

        def on_read_ip_range(start, end):
            """Add ip range to blocklist"""
            # log.trace('Adding ip range %s - %s to ipfilter as blocked', start, end)
            self.blocklist.add_rule(start.address, end.address, BLOCK_RANGE)
            self.num_blocked += 1

        def on_finish_read(result):
            """Add any whitelisted IP's and add the blocklist to session"""
            # White listing happens last because the last rules added have
            # priority
            log.info('Added %d ranges to ipfilter as blocked', self.num_blocked)
            for ip in self.config['whitelisted']:
                ip = IP.parse(ip)
                self.blocklist.add_rule(ip.address, ip.address, ALLOW_RANGE)
                self.num_whited += 1
                log.trace('Added %s to the ipfiler as white-listed', ip.address)
            log.info('Added %d ranges to ipfilter as white-listed', self.num_whited)
            self.core.session.set_ip_filter(self.blocklist)
            return result

        # TODO: double check logic
        if self.up_to_date and self.has_imported:
            log.debug('Latest blocklist is already imported')
            return defer.succeed(blocklist)

        self.is_importing = True
        self.num_blocked = 0
        self.num_whited = 0
        self.blocklist = self.core.session.get_ip_filter()

        if not blocklist:
            blocklist = self.filename

        if not self.reader:
            self.auto_detect(blocklist)
            self.auto_detected = True

        def on_reader_failure(failure):
            log.error('Failed to read!!!!!!')
            log.exception(failure)

        log.debug('Importing using reader: %s', self.reader)
        log.debug(
            'Reader type: %s compression: %s',
            self.config['list_type'],
            self.config['list_compression'],
        )
        log.debug('Clearing current ip filtering')
        # self.blocklist.add_rule('0.0.0.0', '255.255.255.255', ALLOW_RANGE)
        d = threads.deferToThread(self.reader(blocklist).read, on_read_ip_range)
        d.addCallback(on_finish_read).addErrback(on_reader_failure)

        return d

    def on_import_complete(self, blocklist):
        """Runs any import clean up functions.

        Args:
            blocklist (str): Path of blocklist.

        Returns:
            Deferred: A Deferred that fires when clean up is done.

        """
        log.trace('on_import_list_complete')
        d = blocklist
        self.is_importing = False
        self.has_imported = True
        log.debug('Blocklist import complete!')
        cache = deluge.configmanager.get_config_dir('blocklist.cache')
        if blocklist != cache:
            if self.is_url:
                log.debug('Moving %s to %s', blocklist, cache)
                d = threads.deferToThread(shutil.move, blocklist, cache)
            else:
                log.debug('Copying %s to %s', blocklist, cache)
                d = threads.deferToThread(shutil.copy, blocklist, cache)
        return d

    def on_import_error(self, f):
        """Recovers from import error.

        Args:
            f (Failure): Failure that occurred.

        Returns:
            Deferred or Failure: A Deferred if recovery was possible else original Failure.

        """
        log.trace('on_import_error: %s', f)
        d = f
        self.is_importing = False
        try_again = False
        cache = deluge.configmanager.get_config_dir('blocklist.cache')

        if f.check(ReaderParseError) and not self.auto_detected:
            # Invalid / corrupt list, let's detect it
            log.warning('Invalid / corrupt blocklist')
            self.reader = None
            blocklist = None
            try_again = True
        elif self.filename != cache and os.path.exists(cache):
            # If we have a backup and we haven't already used it
            log.warning('Error reading blocklist: %s', f.getErrorMessage())
            blocklist = cache
            try_again = True

        if try_again:
            d = self.import_list(blocklist)
            d.addCallbacks(self.on_import_complete, self.on_import_error)

        return d

    def auto_detect(self, blocklist):
        """Attempts to auto-detect the blocklist type.

        Args:
            blocklist (str): Path of blocklist.

        Raises:
            UnknownFormatError: If the format cannot be detected.

        """
        self.config['list_compression'] = detect_compression(blocklist)
        self.config['list_type'] = detect_format(
            blocklist, self.config['list_compression']
        )
        log.debug(
            'Auto-detected type: %s compression: %s',
            self.config['list_type'],
            self.config['list_compression'],
        )
        if not self.config['list_type']:
            self.config['list_compression'] = ''
            raise UnknownFormatError
        else:
            self.reader = create_reader(
                self.config['list_type'], self.config['list_compression']
            )

    def pause_session(self):
        self.need_to_resume_session = not self.core.session.is_paused()
        self.core.pause_session()

    def resume_session(self, result):
        self.core.resume_session()
        self.need_to_resume_session = False
        return result



#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import os

import pytest
from twisted.internet.utils import getProcessOutputAndValue

import deluge.component as component
import deluge.ui.web.server
from deluge import configmanager
from deluge.common import windows_check
from deluge.conftest import BaseTestCase
from deluge.ui.web.server import DelugeWeb

from .common import get_test_data_file
from .common_web import WebServerTestBase
from .daemon_base import DaemonBase

SECURITY_TESTS = bool(os.getenv('SECURITY_TESTS', False))


# TODO: This whole module has not been tested since migrating tests fully to pytest
class SecurityBaseTestCase:
    @pytest.fixture(autouse=True)
    def setvars(self):
        self.home_dir = os.path.expanduser('~')
        self.port = 8112

    def _run_test(self, test):
        d = getProcessOutputAndValue(
            'bash',
            [
                get_test_data_file('testssl.sh'),
                '--quiet',
                '--nodns',
                'none',
                '--color',
                '0',
                test,
                '127.0.0.1:%d' % self.port,
            ],
        )

        def on_result(results):

            if test == '-e':
                results = results[0].split(b'\n')[7:-6]
                assert len(results) > 3
            else:
                assert b'OK' in results[0]
                assert b'NOT ok' not in results[0]

        d.addCallback(on_result)
        return d

    def test_secured_webserver_protocol(self):
        return self._run_test('-p')

    def test_secured_webserver_standard_ciphers(self):
        return self._run_test('-s')

    def test_secured_webserver_heartbleed_vulnerability(self):
        return self._run_test('-H')

    def test_secured_webserver_css_injection_vulnerability(self):
        return self._run_test('-I')

    def test_secured_webserver_renegotiation_vulnerabilities(self):
        return self._run_test('-R')

    def test_secured_webserver_crime_vulnerability(self):
        return self._run_test('-C')

    def test_secured_webserver_poodle_vulnerability(self):
        return self._run_test('-O')

    def test_secured_webserver_tls_fallback_scsv_mitigation_vulnerability(self):
        return self._run_test('-Z')

    def test_secured_webserver_sweet32_vulnerability(self):
        return self._run_test('-W')

    def test_secured_webserver_beast_vulnerability(self):
        return self._run_test('-A')

    def test_secured_webserver_lucky13_vulnerability(self):
        return self._run_test('-L')

    def test_secured_webserver_freak_vulnerability(self):
        return self._run_test('-F')

    def test_secured_webserver_logjam_vulnerability(self):
        return self._run_test('-J')

    def test_secured_webserver_drown_vulnerability(self):
        return self._run_test('-D')

    def test_secured_webserver_forward_secrecy_settings(self):
        return self._run_test('-f')

    def test_secured_webserver_rc4_ciphers(self):
        return self._run_test('-4')

    def test_secured_webserver_preference(self):
        return self._run_test('-P')

    def test_secured_webserver_ciphers(self):
        return self._run_test('-e')


@pytest.mark.skipif(windows_check(), reason='windows cannot run .sh files')
@pytest.mark.skipif(not SECURITY_TESTS, reason='skipping security tests')
@pytest.mark.security
class TestDaemonSecurity(BaseTestCase, DaemonBase, SecurityBaseTestCase):
    def set_up(self):
        d = self.common_set_up()
        self.port = self.listen_port
        d.addCallback(self.start_core)
        d.addErrback(self.terminate_core)
        return d

    def tear_down(self):
        d = component.shutdown()
        d.addCallback(self.terminate_core)
        return d


@pytest.mark.skipif(windows_check(), reason='windows cannot run .sh files')
@pytest.mark.skipif(not SECURITY_TESTS, reason='skipping security tests')
@pytest.mark.security
class TestWebUISecurity(WebServerTestBase, SecurityBaseTestCase):
    def start_webapi(self, arg):
        self.port = self.webserver_listen_port = 8999

        config_defaults = deluge.ui.web.server.CONFIG_DEFAULTS.copy()
        config_defaults['port'] = self.webserver_listen_port
        config_defaults['https'] = True
        self.config = configmanager.ConfigManager('web.conf', config_defaults)

        self.deluge_web = DelugeWeb(daemon=False)

        host = list(self.deluge_web.web_api.hostlist.config['hosts'][0])
        host[2] = self.listen_port
        self.deluge_web.web_api.hostlist.config['hosts'][0] = tuple(host)
        self.host_id = host[0]
        self.deluge_web.start()

    def test_secured_webserver_headers(self):
        return self._run_test('-h')

    def test_secured_webserver_breach_vulnerability(self):
        return self._run_test('-B')

    def test_secured_webserver_ticketbleed_vulnerability(self):
        return self._run_test('-T')



# vim: sw=4 ts=4 fenc=utf-8 et
# ==============================================================================
# Copyright © 2009-2010 UfSoft.org - Pedro Algarvio <pedro@algarvio.me>
#
# License: BSD - Please view the LICENSE file for additional information.
# ==============================================================================

import logging

from twisted.internet import task

from deluge import component
from deluge.event import DelugeEvent

log = logging.getLogger(__name__)


class FooEvent(DelugeEvent):
    """foo Event"""


class CustomEvent(DelugeEvent):
    """Just a custom event to test"""


class TestEmailNotifications(component.Component):
    def __init__(self, imp):
        component.Component.__init__(self, self.__class__.__name__, 5)
        self.__imp = imp
        self.lc = task.LoopingCall(self.update)
        self.n = 1
        self.events = [FooEvent(), CustomEvent()]
        self.events_classes = []

    def enable(self):
        log.debug('\n\nEnabling %s', self.__class__.__name__)
        for event in self.events:
            if self.__imp == 'core':
                # component.get('CorePlugin.Notifications').register_custom_email_notification(
                component.get('Notifications').register_custom_email_notification(
                    event.__class__.__name__, self.custom_email_message_provider
                )
            elif self.__imp == 'gtk':
                notifications_component = component.get('Notifications')
                notifications_component.register_custom_popup_notification(
                    event.__class__.__name__, self.custom_popup_message_provider
                )
                notifications_component.register_custom_blink_notification(
                    event.__class__.__name__, self.custom_blink_message_provider
                )
                notifications_component.register_custom_sound_notification(
                    event.__class__.__name__, self.custom_sound_message_provider
                )

        self.lc.start(60, False)

    def disable(self):
        log.debug('\n\nDisabling %s', self.__class__.__name__)
        self.lc.stop()

    def update(self):
        if self.__imp == 'core':
            log.debug('\n\nUpdating %s', self.__class__.__name__)
            self.events.append(self.events.pop(0))  # Re-Queue
            self.n += 1
            component.get('EventManager').emit(self.events[0])

    def custom_email_message_provider(self, *evt_args, **evt_kwargs):
        log.debug('Running custom email message provider: %s %s', evt_args, evt_kwargs)
        subject = f'{self.events[0].__class__.__name__} Email Subject: {self.n}'
        message = f'{self.events[0].__class__.__name__} Email Message: {self.n}'
        return subject, message

    def custom_popup_message_provider(self, *evt_args, **evt_kwargs):
        log.debug('Running custom popup message provider: %s %s', evt_args, evt_kwargs)
        title = f'{self.events[0].__class__.__name__} Popup Title: {self.n}'
        message = f'{self.events[0].__class__.__name__} Popup Message: {self.n}'
        return title, message

    def custom_blink_message_provider(self, *evt_args, **evt_kwargs):
        log.debug('Running custom blink message provider: %s %s', evt_args, evt_kwargs)
        return True

    def custom_sound_message_provider(self, *evt_args, **evt_kwargs):
        log.debug('Running custom sound message provider: %s %s', evt_args, evt_kwargs)
        return ''
